"""This module implements some utilities for audio data augmentation with torch_audiomentations."""

from typing import List, Optional

import torch

from senselab.audio.data_structures import (
    Audio,
    batch_audios,
    unbatch_audios,
)
from senselab.utils.data_structures import DeviceType, _select_device_and_dtype

try:
    from torch_audiomentations import Compose

    TORCH_AUDIOMENTATIONS_AVAILABLE = True
except ModuleNotFoundError:
    TORCH_AUDIOMENTATIONS_AVAILABLE = False


def augment_audios_with_torch_audiomentations(
    audios: List[Audio], augmentation: "Compose", device: Optional[DeviceType] = None
) -> List[Audio]:
    """Augments all provided audios with a given augmentation, either individually or all batched together.

    Augment all audios with a user defined augmentation that can be a composition of multiple augmentations.
    This augmentation is either performed on each audio individually (using pydra for optimization)
    or all of the audios provided are batched together and run at once.
    It uses torch_audiomentations. If batching, all audios must have the same sampling rate.

    Args:
        audios: List of Audios whose data will be augmented with the given augmentations
        augmentation: A Composition of augmentations to run on each audio (uses torch-audiomentations), should have its
            output_type set to "dict"
        device: The device to use for augmenting. If the chosen device
            is CUDA then the audios are all batched together, so for optimal performance, batching should
            be done by passing a batch_size worth of audios ar a time.
            Default is None, which will select the device automatically.

    Returns:
        List of audios that has passed the all of input audios through the provided augmentation. This does
            not necessarily mean that the augmentation has been run on every audio. For more information,
            see the torch-audiomentations documentation.
    """
    if not TORCH_AUDIOMENTATIONS_AVAILABLE:
        raise ModuleNotFoundError(
            "`torch-audiomentations` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`"
        )

    augmentation.output_type = "dict"
    device_type, _ = _select_device_and_dtype(
        user_preference=device, compatible_devices=[DeviceType.CUDA, DeviceType.CPU]
    )
    if device_type == DeviceType.CPU:
        '''
        # The commented code is for parallelizing the augmentation using pydra
        # Due to some issues with pydra, this is disabled for now
        import pydra

        def _augment_single_audio(audio: Audio, augmentation: Compose):  # noqa: ANN202
            """Augments a single audio with torch-audiomentations.

            Args:
                audio: The audio to be augmented.
                augmentation: The torch-audiomentations augmentation to be applied.

            Returns:
                The augmented audio. The returned data type is not explicitly specified
                    in the function signature because it would brake pydra.
            """
            augmented_waveform = augmentation(audio.waveform.unsqueeze(0), sample_rate=audio.sampling_rate).samples
            return Audio(
                waveform=torch.squeeze(augmented_waveform),
                sampling_rate=audio.sampling_rate,
                metadata=audio.metadata.copy(),
            )

        _augment_single_audio_pt = pydra.mark.task(_augment_single_audio)

        # Create the workflow
        wf = pydra.Workflow(name="audio_augmentation_wf", input_spec=["y"])
        wf.split("y", y=audios)
        wf.add(_augment_single_audio_pt(name="augment_audio", audio=wf.lzin.y, augmentation=augmentation))
        wf.set_output([("augmented_audio", wf.augment_audio.lzout.out)])

        # Execute the workflow
        with pydra.Submitter(plugin="cf") as submitter:
            submitter(wf)
        outputs = wf.result()
        return [out.output.augmented_audio for out in outputs]
        '''
        new_audios = []
        for audio in audios:
            audio_to_augment = audio.waveform.unsqueeze(0)
            augmented_audio = augmentation(audio_to_augment, sample_rate=audio.sampling_rate).samples
            new_audios.append(
                Audio(
                    waveform=torch.squeeze(augmented_audio),
                    sampling_rate=audio.sampling_rate,
                    metadata=audio.metadata.copy(),
                )
            )
        return new_audios
    else:
        batched_audios, sampling_rates, metadatas = batch_audios(audios)

        batched_audios = batched_audios.to(device=torch.device(device_type.value))
        sampling_rate = sampling_rates[0] if isinstance(sampling_rates, List) else sampling_rates
        augmented_audio = augmentation(batched_audios, sample_rate=sampling_rate).samples

        augmented_audio = augmented_audio.detach().cpu()
        return unbatch_audios(augmented_audio, sampling_rates, metadatas)



"""This module contains functions for plotting audio-related data."""

from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import torch
from IPython.display import Audio, display

from senselab.audio.data_structures import Audio as AudioData
from senselab.audio.tasks.features_extraction.torchaudio import (
    extract_mel_spectrogram_from_audios,
    extract_spectrogram_from_audios,
)
from senselab.utils.data_structures import logger


def plot_waveform(audio: AudioData, title: str = "Waveform", fast: bool = False) -> None:
    """Plots the waveform of an Audio object.

    Args:
        audio (AudioData): An instance of Audio containing waveform data and sampling rate.
        title (str): Title of the plot.
        fast (bool): If True, plots a downsampled version for a faster but less detailed view.

    Todo:
        - Add option to save the plot
        - Add option to choose the size of the Figure
    """
    waveform = audio.waveform
    sample_rate = audio.sampling_rate

    if fast:
        # Downsampling the waveform and sample rate for a quicker plot
        waveform = waveform[::10]

    num_channels, num_frames = waveform.shape
    time_axis = torch.linspace(0, num_frames / sample_rate, num_frames)

    figure, axes = plt.subplots(num_channels, 1, figsize=(12, num_channels * 2))
    if num_channels == 1:
        axes = [axes]  # Ensure axes is iterable
    for c, ax in enumerate(axes):
        ax.plot(time_axis, waveform[c].numpy(), linewidth=1)
        ax.set_ylabel(f"Channel {c + 1}")
        ax.grid(True)

    figure.suptitle(title)
    plt.xlabel("Time [s]")
    plt.show(block=False)


def plot_specgram(audio: AudioData, mel_scale: bool = False, title: str = "Spectrogram", **spect_kwargs: Any) -> None:  # noqa : ANN401
    """Plots the spectrogram of an Audio object.

    Args:
        audio: (AudioData): An instance of Audio containing waveform data and sampling rate.
        mel_scale (bool): Whether to plot a mel spectrogram or a regular spectrogram.
        title (str): Title of the spectrogram plot.
        **spect_kwargs: Additional keyword arguments to pass to the spectrogram function.

    Todo:
        - Add option to save the plot
        - Add option to choose the size of the Figure
    """

    def _power_to_db(
        spectrogram: np.ndarray, ref: float = 1.0, amin: float = 1e-10, top_db: float = 80.0
    ) -> np.ndarray:
        """Converts a power spectrogram (amplitude squared) to decibel (dB) units.

        Args:
            spectrogram (np.ndarray): Power spectrogram to convert.
            ref (float): Reference power. Default is 1.0.
            amin (float): Minimum power. Default is 1e-10.
            top_db (float): Minimum decibel. Default is 80.0.

        Returns:
            np.ndarray: Decibel spectrogram.
        """
        S = np.asarray(spectrogram)

        if amin <= 0:
            raise ValueError("amin must be strictly positive")

        if np.issubdtype(S.dtype, np.complexfloating):
            logger.warning(
                "_power_to_db was called on complex input so phase "
                "information will be discarded. To suppress this warning, "
                "call power_to_db(np.abs(D)**2) instead.",
                stacklevel=2,
            )
            magnitude = np.abs(S)
        else:
            magnitude = S

        if callable(ref):
            # User supplied a function to calculate reference power
            ref_value = ref(magnitude)
        else:
            ref_value = np.abs(ref)

        log_spec: np.ndarray = 10.0 * np.log10(np.maximum(amin, magnitude))
        log_spec -= 10.0 * np.log10(np.maximum(amin, ref_value))

        if top_db is not None:
            if top_db < 0:
                raise ValueError("top_db must be non-negative")
            log_spec = np.maximum(log_spec, log_spec.max() - top_db)

        return log_spec

    # Extract the spectrogram
    if mel_scale:
        spectrogram = extract_mel_spectrogram_from_audios([audio], **spect_kwargs)[0]["mel_spectrogram"]
        y_axis_label = "Mel Frequency"
    else:
        spectrogram = extract_spectrogram_from_audios([audio], **spect_kwargs)[0]["spectrogram"]
        y_axis_label = "Frequency [Hz]"

    if spectrogram.dim() != 2:
        raise ValueError("Spectrogram must be a 2D tensor.")

    # Determine time and frequency scale
    num_frames = spectrogram.size(1)
    num_freq_bins = spectrogram.size(0)

    # Time axis in seconds
    time_axis = (audio.waveform.size(-1) / audio.sampling_rate) * (torch.arange(0, num_frames).float() / num_frames)

    # Frequency axis in Hz (for non-mel spectrograms)
    if mel_scale:
        freq_axis = torch.arange(num_freq_bins)  # For mel spectrogram, keep the bins as discrete values
    else:
        freq_axis = torch.linspace(0, audio.sampling_rate / 2, num_freq_bins)

    plt.figure(figsize=(10, 4))
    plt.imshow(
        _power_to_db(spectrogram.numpy()),
        aspect="auto",
        origin="lower",
        extent=(float(time_axis[0]), float(time_axis[-1]), float(freq_axis[0]), float(freq_axis[-1])),
        cmap="viridis",
    )
    plt.colorbar(label="Magnitude (dB)")
    plt.title(title)
    plt.ylabel(y_axis_label)
    plt.xlabel("Time [Sec]")
    plt.show(block=False)


def play_audio(audio: AudioData) -> None:
    """Plays an audio file.

    Args:
        audio (AudioData): An instance of Audio containing waveform data and sampling rate.

    Raises:
        ValueError: If the number of channels is more than 2.
    """
    waveform = audio.waveform.numpy()
    sample_rate = audio.sampling_rate

    num_channels = waveform.shape[0]
    if num_channels == 1:
        display(Audio(waveform[0], rate=sample_rate))
    elif num_channels == 2:
        display(Audio((waveform[0], waveform[1]), rate=sample_rate))
    else:
        raise ValueError("Waveform with more than 2 channels are not supported.")



"""This module provides the implementation of torchaudio utilities for audio features extraction."""

import os
from typing import Any, Dict, List, Optional

import numpy as np
import pydra
import torch

try:
    import torchaudio

    TORCHAUDIO_AVAILABLE = True
except ModuleNotFoundError:
    TORCHAUDIO_AVAILABLE = False

from senselab.audio.data_structures import Audio


def extract_spectrogram_from_audios(
    audios: List[Audio],
    n_fft: int = 1024,
    win_length: Optional[int] = None,
    hop_length: Optional[int] = None,
) -> List[Dict[str, torch.Tensor]]:
    """Extract spectrograms from a list of audio objects.

    Args:
        audios (List[Audio]): List of Audio objects.
        n_fft (int): Size of FFT, creates n_fft // 2 + 1 bins. Default is 1024.
        win_length (int): Window size. Default is None, using n_fft.
        hop_length (int): Length of hop between STFT windows. Default is None, using win_length // 2.

    Returns:
        List[Dict[str, torch.Tensor]]: List of Dict objects containing spectrograms.
    """
    if not TORCHAUDIO_AVAILABLE:
        raise ModuleNotFoundError(
            "`torchaudio` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )

    if win_length is None:
        win_length = n_fft
    if hop_length is None:
        hop_length = win_length // 2
    spectrogram = torchaudio.transforms.Spectrogram(
        n_fft=n_fft,
        win_length=win_length,
        hop_length=hop_length,
    )
    spectrograms = []
    for audio in audios:
        try:
            spectrograms.append({"spectrogram": spectrogram(audio.waveform).squeeze(0)})
        except RuntimeError:
            spectrograms.append({"spectrogram": np.nan})
    return spectrograms


def extract_mel_spectrogram_from_audios(
    audios: List[Audio],
    n_fft: Optional[int] = 1024,
    win_length: Optional[int] = None,
    hop_length: Optional[int] = None,
    n_mels: int = 128,
) -> List[Dict[str, torch.Tensor]]:
    """Extract mel spectrograms from a list of audio objects.

    Args:
        audios (List[Audio]): List of Audio objects.
        n_fft (int): Size of FFT, creates n_fft // 2 + 1 bins. Default is 1024.
        win_length (int): Window size. Default is None, using n_fft.
        hop_length (int): Length of hop between STFT windows. Default is None, using win_length // 2.
        n_mels (int): Number of mel filter banks. Default is 128.

    Returns:
        List[Dict[str, torch.Tensor]]: List of Dict objects containing mel spectrograms.
    """
    if not TORCHAUDIO_AVAILABLE:
        raise ModuleNotFoundError(
            "`torchaudio` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )

    if win_length is None:
        win_length = n_fft
    if hop_length is None:
        if win_length is not None:
            hop_length = win_length // 2
        else:
            raise ValueError("win_length cannot be None")
    mel_spectrograms = []
    for audio in audios:
        try:
            mel_spectrogram = torchaudio.transforms.MelSpectrogram(
                sample_rate=audio.sampling_rate,
                n_fft=n_fft,
                win_length=win_length,
                hop_length=hop_length,
                n_mels=n_mels,
            )(audio.waveform)
            mel_spectrograms.append({"mel_spectrogram": mel_spectrogram.squeeze(0)})
        except RuntimeError:
            mel_spectrograms.append({"mel_spectrogram": np.nan})
    return mel_spectrograms


def extract_mfcc_from_audios(
    audios: List[Audio],
    n_mfcc: int = 40,
    n_ftt: Optional[int] = 400,
    win_length: Optional[int] = None,
    hop_length: Optional[int] = None,
    n_mels: int = 128,
) -> List[Dict[str, torch.Tensor]]:
    """Extract MFCCs from a list of audio objects.

    Args:
        audios (List[Audio]): List of Audio objects.
        n_mfcc (int): Number of MFCCs to return. Default is 40.
        n_ftt (int): Size of FFT, creates n_ftt // 2 + 1 bins. Default is 400.
        win_length (int): Window size. Default is None, using n_ftt.
        hop_length (int): Length of hop between STFT windows. Default is None, using win_length // 2.
        n_mels (int): Number of mel filter banks. Default is 128.

    Returns:
        List[Dict[str, torch.Tensor]]: List of Dict objects containing MFCCs.
    """
    if not TORCHAUDIO_AVAILABLE:
        raise ModuleNotFoundError(
            "`torchaudio` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )

    if win_length is None:
        win_length = n_ftt
    if hop_length is None:
        if win_length is not None:
            hop_length = win_length // 2
        else:
            raise ValueError("win_length cannot be None")
    mfccs = []
    for audio in audios:
        try:
            mfcc_transform = torchaudio.transforms.MFCC(
                sample_rate=audio.sampling_rate,
                n_mfcc=n_mfcc,
                melkwargs={"n_fft": n_ftt, "win_length": win_length, "hop_length": hop_length, "n_mels": n_mels},
            )
            mfccs.append({"mfcc": mfcc_transform(audio.waveform).squeeze(0)})
        except RuntimeError:
            mfccs.append({"mfcc": np.nan})
    return mfccs


def extract_mel_filter_bank_from_audios(
    audios: List[Audio],
    n_mels: int = 128,
    n_fft: int = 1024,
    win_length: Optional[int] = None,
    hop_length: Optional[int] = None,
) -> List[Dict[str, torch.Tensor]]:
    """Extract mel filter bank from a list of audio objects.

    Args:
        audios (List[Audio]): List of Audio objects.
        n_mels (int): Number of mel filter banks. Default is 128.
        n_fft (int): Size of FFT, creates n_fft // 2 + 1 bins. Default is 1024.
        win_length (int): Window size. Default is None, using n_fft.
        hop_length (int): Length of hop between STFT windows. Default is None, using win_length // 2.

    Returns:
        List[Dict[str, torch.Tensor]]: List of Dict objects containing mel filter banks.
    """
    if not TORCHAUDIO_AVAILABLE:
        raise ModuleNotFoundError(
            "`torchaudio` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )

    if win_length is None:
        win_length = n_fft
    if hop_length is None:
        hop_length = win_length // 2
    n_stft = n_fft // 2 + 1

    spectrograms = extract_spectrogram_from_audios(audios, n_fft, win_length, hop_length)

    mel_filter_banks = []
    for i, audio in enumerate(audios):
        try:
            melscale_transform = torchaudio.transforms.MelScale(
                sample_rate=audio.sampling_rate, n_mels=n_mels, n_stft=n_stft
            )
            mel_filter_banks.append({"mel_filter_bank": melscale_transform(spectrograms[i]["spectrogram"]).squeeze(0)})
        except RuntimeError:
            mel_filter_banks.append({"mel_filter_bank": np.nan})
    return mel_filter_banks


def extract_mel_filter_bank_from_spectrograms(
    spectrograms: List[Dict[str, torch.Tensor]],
    sampling_rate: int,
    n_mels: int = 128,
) -> List[Dict[str, torch.Tensor]]:
    """Extract mel filter bank from a list of audio objects.

    Args:
        spectrograms (List[torch.Tensor]): List of spectrograms.
        sampling_rate (int): Sampling rate of the audio.
        n_mels (int): Number of mel filter banks. Default is 128.

    Returns:
        List[Dict[str, torch.Tensor]]: List of Dict objects containing mel filter banks.
    """
    if not TORCHAUDIO_AVAILABLE:
        raise ModuleNotFoundError(
            "`torchaudio` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )

    mel_filter_banks = []
    for spectrogram in spectrograms:
        try:
            melscale_transform = torchaudio.transforms.MelScale(
                sample_rate=sampling_rate, n_mels=n_mels, n_stft=spectrogram["spectrogram"].shape[0]
            )
            mel_filter_banks.append({"mel_filter_bank": melscale_transform(spectrogram["spectrogram"]).squeeze(0)})
        except RuntimeError:
            mel_filter_banks.append({"mel_filter_bank": np.nan})
    return mel_filter_banks


def extract_pitch_from_audios(
    audios: List[Audio], freq_low: int = 80, freq_high: int = 500
) -> List[Dict[str, torch.Tensor]]:
    """Extract pitch from a list of audio objects.

    Pitch is detected using the detect_pitch_frequency function from torchaudio.
    It is implemented using normalized cross-correlation function and median smoothing.

    Args:
        audios (List[Audio]): List of Audio objects.
        freq_low (int): Lowest frequency that can be detected (Hz). Should be bigger than 0.
            (Default is 80).
        freq_high (int): Highest frequency that can be detected (Hz).
            (Default is 500).

    Returns:
        List[Dict[str, torch.Tensor]]: List of Dict objects containing pitches.
    """
    if not TORCHAUDIO_AVAILABLE:
        raise ModuleNotFoundError(
            "`torchaudio` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )

    if freq_low <= 0:
        raise ValueError("freq_low should be bigger than 0")

    pitches = []
    for audio in audios:
        try:
            pitches.append(
                {
                    "pitch": torchaudio.functional.detect_pitch_frequency(
                        audio.waveform, sample_rate=audio.sampling_rate, freq_low=freq_low, freq_high=freq_high
                    ).squeeze(0)
                }
            )
        except RuntimeError:
            pitches.append({"pitch": torch.tensor(torch.nan)})
    return pitches


def extract_torchaudio_features_from_audios(
    audios: List[Audio],
    freq_low: int = 80,
    freq_high: int = 500,
    n_fft: int = 1024,
    n_mels: int = 128,
    n_mfcc: int = 40,
    win_length: Optional[int] = None,
    hop_length: Optional[int] = None,
    plugin: str = "serial",
    plugin_args: Optional[Dict[str, Any]] = {},
    cache_dir: Optional[str | os.PathLike] = None,
) -> List[Dict[str, Any]]:
    """Extract torchaudio features from a list of audio objects.

    Args:
        audios (List[Audio]): The list of audio objects to extract features from.
        freq_low (int): Lowest frequency that can be detected (Hz). Should be bigger than 0.
            (Default is 80).
        freq_high (int): Highest frequency that can be detected (Hz).
            (Default is 500).
        n_fft (int): Size of FFT, creates n_fft // 2 + 1 bins. Default is 1024.
        n_mels (int): Number of mel filter banks. Default is 128.
        n_mfcc (int): Number of MFCCs. Default is 40.
        win_length (int): Window size. Default is None, using n_fft.
        hop_length (int): Length of hop between STFT windows. Default is None, using win_length // 2.
        plugin (str): The plugin to use. Default is "serial".
        plugin_args (Optional[Dict[str, Any]]): The arguments to pass to the plugin. Default is {}.
        cache_dir (Optional[str | os.PathLike]): The directory to cache the results. Default is None.

    Returns:
        List[Dict[str, Any]]: The list of feature dictionaries for each audio.
    """
    extract_pitch_from_audios_pt = pydra.mark.task(extract_pitch_from_audios)
    extract_mel_filter_bank_from_spectrograms_pt = pydra.mark.task(extract_mel_filter_bank_from_spectrograms)
    extract_mfcc_from_audios_pt = pydra.mark.task(extract_mfcc_from_audios)
    extract_mel_spectrogram_from_audios_pt = pydra.mark.task(extract_mel_spectrogram_from_audios)
    extract_spectrogram_from_audios_pt = pydra.mark.task(extract_spectrogram_from_audios)

    def _extract_sampling_rate(audios: List[Audio]) -> int:
        """Extract the sampling rate from an Audio object."""
        return audios[0].sampling_rate

    _extract_sampling_rate_pt = pydra.mark.task(_extract_sampling_rate)

    formatted_audios = [[audio] for audio in audios]
    wf = pydra.Workflow(name="wf", input_spec=["x"], cache_dir=cache_dir)
    wf.split("x", x=formatted_audios)
    wf.add(_extract_sampling_rate_pt(name="_extract_sampling_rate_pt", audios=wf.lzin.x))
    wf.add(
        extract_pitch_from_audios_pt(
            name="extract_pitch_from_audios_pt", audios=wf.lzin.x, freq_low=freq_low, freq_high=freq_high
        )
    )
    wf.add(
        extract_spectrogram_from_audios_pt(
            name="extract_spectrogram_from_audios_pt",
            audios=wf.lzin.x,
            n_fft=n_fft,
            win_length=win_length,
            hop_length=hop_length,
        )
    )
    wf.add(
        extract_mel_spectrogram_from_audios_pt(
            name="extract_mel_spectrogram_from_audios_pt",
            audios=wf.lzin.x,
            n_mels=n_mels,
            n_fft=n_fft,
            win_length=win_length,
            hop_length=hop_length,
        )
    )
    wf.add(
        extract_mel_filter_bank_from_spectrograms_pt(
            name="extract_mel_filter_bank_from_spectrograms_pt",
            spectrograms=wf.extract_spectrogram_from_audios_pt.lzout.out,
            sampling_rate=wf._extract_sampling_rate_pt.lzout.out,
            n_mels=n_mels,
            n_fft=n_fft,
        )
    )
    wf.add(
        extract_mfcc_from_audios_pt(
            name="extract_mfcc_from_audios_pt",
            audios=wf.lzin.x,
            n_mfcc=n_mfcc,
            n_fft=n_fft,
            n_mels=n_mels,
            win_length=win_length,
            hop_length=hop_length,
        )
    )

    # setting multiple workflow outputs
    wf.set_output(
        [
            ("pitch_out", wf.extract_pitch_from_audios_pt.lzout.out),
            ("mel_filter_bank_out", wf.extract_mel_filter_bank_from_spectrograms_pt.lzout.out),
            ("mfcc_out", wf.extract_mfcc_from_audios_pt.lzout.out),
            ("mel_spectrogram_out", wf.extract_mel_spectrogram_from_audios_pt.lzout.out),
            ("spectrogram_out", wf.extract_spectrogram_from_audios_pt.lzout.out),
        ]
    )

    with pydra.Submitter(plugin=plugin, **plugin_args) as sub:
        sub(wf)

    outputs = wf.result()

    formatted_output: List[Dict[str, Any]] = []
    for output in outputs:
        formatted_output_item = {
            "pitch": output.output.pitch_out[0]["pitch"],
            "mel_filter_bank": output.output.mel_filter_bank_out[0]["mel_filter_bank"],
            "mfcc": output.output.mfcc_out[0]["mfcc"],
            "mel_spectrogram": output.output.mel_spectrogram_out[0]["mel_spectrogram"],
            "spectrogram": output.output.spectrogram_out[0]["spectrogram"],
        }

        formatted_output.append(formatted_output_item)

    return formatted_output



"""Provides functionality for batch processing."""

from typing import Any, List


def batch_list(items: List[Any], n_batches: int) -> List[List[Any]]:
    """Splits a list into a specified number of batches.

    Args:
        items (List[Any]): The list of items to be batched.
        n_batches (int): The number of batches to divide the list into.

    Returns:
        List[List[Any]]: A list of lists, where each sublist is a batch of items.
    """
    if not items:
        return []

    n_items = len(items)
    n_batches = min(n_batches, n_items)  # ensure n_batches does not exceed n_items
    batch_size = -(n_items // -n_batches)  # fast ceiling division
    return [items[i : i + batch_size] for i in range(0, n_items, batch_size)]



"""Data structures relevant for managing datasets."""

import math
import uuid
from typing import Dict, List, Union, no_type_check

import torch
from datasets import Audio as HFAudio
from datasets import Dataset, Features, Image, Sequence, Value
from pydantic import BaseModel, Field, ValidationInfo, field_validator
from transformers.image_transforms import to_pil_image

from senselab.audio.data_structures import Audio
from senselab.video.data_structures import Video


class Participant(BaseModel):
    """Data structure for a participant in a dataset.

    Attributes:
        id: The ID of the participant. If not provided, generated using uuid4
        metadata: Dictionary of user specified metadata related to the participant
    """

    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    metadata: Dict = Field(default={})

    @field_validator("id", mode="before")
    def set_id(cls, v: str) -> str:
        """Set the unique id of the participant."""
        return v or str(uuid.uuid4())

    def __eq__(self, other: object) -> bool:
        """Overloads the default BaseModel equality to correctly check that ids are equivalent."""
        if isinstance(other, Participant):
            return self.id == other.id
        return False


class Session(BaseModel):
    """Data structure for a session in a dataset.

    Attributes:
        id: the ID of the session. If not provided, generated using uuid4
        metadata: Dictionary of user specified metadata related to the session
    """

    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    metadata: Dict = Field(default={})

    @field_validator("id", mode="before")
    def set_id(cls, v: str) -> str:
        """Set the unique id of the session."""
        return v or str(uuid.uuid4())

    def __eq__(self, other: object) -> bool:
        """Overloads the default BaseModel equality to correctly check that ids are equivalent."""
        if isinstance(other, Session):
            return self.id == other.id
        return False


class SenselabDataset(BaseModel):
    """Class for maintaining SenseLab datasets and functionalities.

    Maintains collections of Audios, Videos, and metadata for use of the Senselab tools
    and pipelines. Includes the ability to manage Sessions and Participants.

    Attributes:
        audios: List of Audios that are generated based on list of audio filepaths
        videos: List of Videos generated from a list of video filepaths
        metadata: Metadata related to the dataset overall but not necessarily the metadata of
            indivudal audios or videos in the dataset
        sessions: Mapping of Session IDs to Session instances
        participants: Mapping of participant IDs to Participant instances
    """

    participants: Dict[str, Participant] = Field(default_factory=dict)
    sessions: Dict[str, Session] = Field(default_factory=dict)
    audios: List[Audio] = []
    videos: List[Video] = []
    metadata: Dict = Field(default={})

    @field_validator("audios", mode="before")
    @classmethod
    def generate_audios_from_filepaths(cls, v: Union[List[str], List[Audio]], _: ValidationInfo) -> List[Audio]:
        """Generate the audios in the dataset from a list of audio filepaths.

        Generates the audios in the dataset by taking in a list of audio filepaths
        or a list of Audios

        Args:
            v: Input for audios attribute that we're validating by generating the Audios if filepaths
                are provided or just the list of Audios if pre-generated and passed in

        Returns:
            List of Audios that instantiates the audios attribute in the dataset
        """
        audio_list = []
        if len(v) == 0:
            return []
        else:
            for audio in v:
                if isinstance(audio, Audio):
                    audio_list.append(audio)
                else:
                    audio_list.append(Audio(filepath=audio))
        return audio_list

    @field_validator("videos", mode="before")
    @classmethod
    def generate_videos_from_filepaths(cls, v: Union[List[str], List[Video]], _: ValidationInfo) -> List[Video]:
        """Generate the videos in the dataset from a list of video filepaths.

        Generates the videos in the dataset by taking in a list of video filepaths
        or a list of Videos

        Args:
            v: Input for videos attribute that we're validating by generating the Videos if filepaths
                are provided or just the list of Videos if pre-generated and passed in

        Returns:
            List of Videos that instantiates the videos attribute in the dataset
        """
        video_list = []
        if len(v) == 0:
            return []
        else:
            for video in v:
                if isinstance(video, Video):
                    video_list.append(video)
                elif isinstance(video, str):
                    video_list.append(Video(filepath=video))

                else:
                    raise ValueError("Unsupported video list")
        return video_list

    @classmethod
    @no_type_check
    def create_bids_dataset(cls, bids_root_filepath: str) -> "SenselabDataset":
        """Create a dataset from a BIDS organized directory.

        Creates a new dataset based off of a BIDS directory structure as defined at
        https://sensein.group/biometrics-book/updated_bids.html
        """
        pass

    # TODO Decide if this method and audio_merge_from_pydra_task should be defined elsewhere, like in
    #     a Pydra helper class
    def create_audio_split_for_pydra_task(self, batch_size: int = 1) -> List[List[Audio]]:
        """Splits the audio data for Pydra tasks.

        Creates a split of the audio data that can be used for creating individual Pydra tasks using
        the .split functionality. Splits the data such that the inputs for a Pydra workflow are either
        optimized for the GPU's batch size or a single Audio per CPU thread.

        Args:
            batch_size: How to batch Audios for a Pydra task; defaults to 1 since CPU won't batch

        Returns:
            List of Lists of Audio where each List of Audios will be an input to a Pydra task.
            Each of the sublists are either of size 1 for CPUs or at most batch_size for GPU optimization.

        Raises:
            ValueError if the batch size is invalid (less than 1)
        """
        if batch_size > 1:
            # Creates batches of at most size batch_size except the last which contains the remainder of audios
            return [
                self.audios[batch_size * i : min(batch_size * (i + 1), len(self.audios))]
                for i in range(math.ceil(len(self.audios) / batch_size))
            ]
        elif batch_size < 1:
            raise ValueError("Batch size must be greater than or equal to 1")
        else:
            return [[audio] for audio in self.audios]

    def audio_merge_from_pydra_task(self, audios_to_merge: List[List[Audio]]) -> None:
        """TODO: Write later.

        Logic Pydra:
        audios: List of audios that want to give to task
        split: List[List[Audios]] -> task List[Audio]
        pydra task(List[Audio]) -> List[Audio]
        merge(List[List[Audio]]) <- might be a wrapped instead of List of lists
        TODO: Figure out what a merge behavior looks like from Pydra
        """
        self.audios = []
        for audio_task_input in audios_to_merge:
            for audio_output in audio_task_input:
                self.audios.append(audio_output)

    def add_participant(self, participant: Participant) -> None:
        """Add a participant to the dataset.

        Adds a new participant to the dataset if they are not already in it.

        Args:
            participant: instance of a Participant that we want to add to the dataset

        Raises:
            ValueError: If the participant ID is already in the dataset, we raise a value error.
              This means that either the ID is non-unique and/or the participant is already in the dataset.
        """
        if participant.id in self.participants:
            raise ValueError(f"Participant with ID {participant.id} already exists.")
        self.participants[participant.id] = participant

    def add_session(self, session: Session) -> None:
        """Add a session to the dataset.

        Adds a new sesszion to the dataset if it is not already in there.

        Args:
            session: instance of a Session that we want to add to the dataset

        Raises:
            ValueError: If the session ID is already in the dataset, we raise a value error.
              This means that either the ID is non-unique and/or the session is already in the dataset.
        """
        if session.id in self.sessions:
            raise ValueError(f"Session with ID {session.id} already exists.")
        self.sessions[session.id] = session

    def get_participants(self) -> List[Participant]:
        """Get the list of participants in the dataset.

        Returns:
            participants (List[Participant]): all of the instances of participants in the dataset
            Warning: The instances are returned as is, so changes to the underlying participants in this
              list will automatically be reflected in the dataset.
        """
        return list(self.participants.values())

    def get_sessions(self) -> List[Session]:
        """Get the list of sessions in the dataset.

        Returns:
            sessions (List[Session]): all of the instances of sessions in the dataset
            Warning: The instances are returned as is, so changes to the underlying sessions in this
              list will automatically be reflected in the dataset.
        """
        return list(self.sessions.values())

    def _get_dict_representation(self) -> Dict:
        """Internal function for generating a dictionary representation of the dataset.

        Returns:
            Generates a dictionary representation of the dataset where the keys are participants,
              sessions, audios, videos, and metadata.
        """
        audio_data: Dict[str, List] = {}
        video_data: Dict[str, List] = {}
        senselab_dict: Dict[str, Union[Dict[str, List], List]] = {
            "participants": [],
            "sessions": [],
            "audios": audio_data,
            "videos": video_data,
            "metadata": self.metadata.copy(),
        }
        participants_data = []
        sessions_data = []

        video_frames_data = []
        video_fps_data = []
        video_metadata = []
        video_audio_data = []
        video_audio_metadata = []

        audio_waveform_data = []
        audio_metadata = []

        for participant in self.get_participants():
            participants_data.append({"id": participant.id, "metadata": participant.metadata.copy()})
        senselab_dict["participants"] = participants_data

        for session in self.get_sessions():
            sessions_data.append({"id": session.id, "metadata": session.metadata.copy()})
        senselab_dict["sessions"] = sessions_data

        for audio in self.audios:
            audio_waveform_data.append(
                {
                    "array": audio.waveform.T,
                    "sampling_rate": audio.sampling_rate,
                }
            )
            audio_metadata.append(audio.metadata.copy())
        audio_data["audio"] = audio_waveform_data
        audio_data["metadata"] = audio_metadata

        for video in self.videos:
            video_frames_data.append({"image": [to_pil_image(frame.numpy()) for frame in list(video.frames)]})
            video_fps_data.append(video.frame_rate)
            video_metadata.append(video.metadata.copy())
            video_audio_data.append(
                None
                if not video.audio
                else {
                    "array": video.audio.waveform.T.to(torch.float32).numpy(),
                    "sampling_rate": video.audio.sampling_rate,
                }
            )
            video_audio_metadata.append(None if not video.audio else video.audio.metadata.copy())

        video_data["frames"] = video_frames_data
        video_data["frame_rate"] = video_fps_data
        video_data["metadata"] = video_metadata
        video_data["audio"] = video_audio_data
        video_data["audio_metadata"] = video_audio_metadata
        # raise ValueError('fuck')
        return senselab_dict

    def convert_senselab_dataset_to_hf_datasets(self) -> Dict[str, Dataset]:
        """Converts Senselab datasets into HuggingFace datasets.

        Returns:
            A dictionary of HuggingFace datasets that represent the underlying Senselab dataset.
              Currently only supports creating HuggingFace datasets for the Audio(s) and Video(s) in
              the SenselabDataset. Videos in HuggingFace are not natively supported, so they are treated
              as Sequences of images with a frame_rate.
        """
        senselab_dict = self._get_dict_representation()

        # print(senselab_dict['videos']['audio'][0])

        features = Features(
            {
                "frames": {"image": Sequence(feature=Image())},
                "frame_rate": Value("float32"),
                "metadata": {},
                "audio": HFAudio(mono=False, sampling_rate=48000),
                "audio_metadata": Value("string"),
            }
        )

        audio_dataset = Dataset.from_dict(senselab_dict["audios"]).cast_column("audio", HFAudio(mono=False))

        video_dataset = Dataset.from_dict(senselab_dict["videos"], features=features)
        # print(video_dataset[0])

        hf_datasets = {}
        hf_datasets["audios"] = audio_dataset
        hf_datasets["videos"] = video_dataset

        # TODO: Create datasets for participants and sessions
        return hf_datasets

    @classmethod
    def convert_hf_dataset_to_senselab_dataset(
        cls, hf_datasets: Dict[str, Dataset], metadata: Dict = {}, transfer_metadata: bool = False
    ) -> "SenselabDataset":
        """Converts HuggingFace dataset to a Senselab dataset.

        Convert HuggingFace dataset(s) to a SenselabDataset where each component of a SenselabDataset
          (e.g. audios, videos, sessions, participants) are stored under different keys in the provided dictionary.

        Args:
            hf_datasets: Dictionary of the different individual components that make up a SenselabDataset. Audios must
              be organized as HuggingFace Audio(s) which is a dictionary with an array attribute, sampling rate, and
              a path. Videos are not natively supported by HuggingFace, so SenselabDataset expects a sequence of
              frames that are each an image, a frame_rate argument, and optionally the associated audio with the
              video.
            metadata: Dictionary of additional dataset level metadata. Differs from metadata inside each HuggingFace
              dataset that is provided and is unaffected by transder_metadata.
            transfer_metadata: Specifies whether to generate metadata from extraneous attributes in the
              HuggingDataset in addition to any specified ina metadata field, otherwise they are ignored.
              Defaults to False.

        Returns:
            The generated SenselabDataset from the provided fields in the hf_datasets dictionary. Currently does not
              support converting sessions or participants.
        """
        audios = []
        videos = []
        sessions: Dict[str, Session] = {}
        participants: Dict[str, Participant] = {}

        if "audios" in hf_datasets:
            audio_dataset = hf_datasets["audios"]
            for audio in audio_dataset:  # Equivalent of running over each row in the Dataset
                audio_metadata = audio["metadata"] if "metadata" in audio else {}
                if transfer_metadata:
                    for feature in audio_dataset.features:
                        if feature == "metadata" or feature == "audio":
                            continue
                        audio_metadata[feature] = audio[feature]
                audios.append(
                    Audio(
                        waveform=audio["audio"]["array"],
                        sampling_rate=audio["audio"]["sampling_rate"],
                        metadata=audio_metadata,
                    )
                )

        if "videos" in hf_datasets:
            video_dataset = hf_datasets["videos"]
            for video in video_dataset:
                video_metadata = video["metadata"] if "metadata" in video else {}
                if transfer_metadata:
                    for feature in video_dataset.features:
                        if (
                            feature == "metadata"
                            or feature == "frames"
                            or feature == "frame_rate"
                            or feature == "audio"
                        ):
                            continue
                        video_metadata[feature] = video[feature]
                videos.append(
                    Video(
                        frames=video["frames"]["image"],
                        frame_rate=video["frame_rate"],
                        metadata=video_metadata,
                        audio=Audio(  # Assumes audio metadata is stored a level higher within the video's metadata
                            waveform=video["audio"]["array"],
                            sampling_rate=video["audio"]["sampling_rate"],
                        )
                        if video["audio"]
                        else None,
                    )
                )
        if "sessions" in hf_datasets:
            pass
        if "participants" in hf_datasets:
            pass

        return SenselabDataset(
            participants=participants, sessions=sessions, audios=audios, videos=videos, metadata=metadata.copy()
        )

    def __eq__(self, other: object) -> bool:
        """Overloads the default BaseModel equality to correctly check that datasets are equivalent."""
        if isinstance(other, SenselabDataset):
            return (
                self.audios == other.audios
                and self.videos == other.videos
                and self.participants == other.participants
                and self.sessions == other.sessions
            )
        return False



"""This module implements some utilities for evaluating a transcription."""

try:
    import jiwer

    JIWER_AVAILABLE = True
except ModuleNotFoundError:
    JIWER_AVAILABLE = False

# TODO: add more metrics which take into account the meaning/intention


def calculate_wer(reference: str, hypothesis: str) -> float:
    """Calculate the Word Error Rate (WER) between the reference and hypothesis.

    Args:
        reference (str): The ground truth text.
        hypothesis (str): The predicted text.

    Returns:
        float: The WER score.

    Examples:
        >>> calculate_wer("hello world", "hello duck")
        0.5
    """
    if not JIWER_AVAILABLE:
        raise ModuleNotFoundError(
            "`jiwer` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )

    return jiwer.wer(reference, hypothesis)


def calculate_mer(reference: str, hypothesis: str) -> float:
    """Calculate the Match Error Rate (MER) between the reference and hypothesis.

    Args:
        reference (str): The ground truth text.
        hypothesis (str): The predicted text.

    Returns:
        float: The MER score.

    Examples:
        >>> calculate_mer("hello world", "hello duck")
        0.5
    """
    if not JIWER_AVAILABLE:
        raise ModuleNotFoundError(
            "`jiwer` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )
    return jiwer.mer(reference, hypothesis)


def calculate_wil(reference: str, hypothesis: str) -> float:
    """Calculate the Word Information Lost (WIL) between the reference and hypothesis.

    Args:
        reference (str): The ground truth text.
        hypothesis (str): The predicted text.

    Returns:
        float: The WIL score.

    Examples:
        >>> calculate_wil("hello world", "hello duck")
        0.75
    """
    if not JIWER_AVAILABLE:
        raise ModuleNotFoundError(
            "`jiwer` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )
    return jiwer.wil(reference, hypothesis)


def calculate_wip(reference: str, hypothesis: str) -> float:
    """Calculate the Word Information Preserved (WIP) between the reference and hypothesis.

    Args:
        reference (str): The ground truth text.
        hypothesis (str): The predicted text.

    Returns:
        float: The WIP score.

    Examples:
        >>> calculate_wip("hello world", "hello duck")
        0.25
    """
    if not JIWER_AVAILABLE:
        raise ModuleNotFoundError(
            "`jiwer` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )
    return jiwer.wip(reference, hypothesis)


def calculate_cer(reference: str, hypothesis: str) -> float:
    """Calculate the Character Error Rate (CER) between the reference and hypothesis.

    Args:
        reference (str): The ground truth text.
        hypothesis (str): The predicted text.

    Returns:
        float: The CER score.

    Examples:
        >>> calculate_cer("hello world", "hello duck")
        0.45454545454545453
    """
    if not JIWER_AVAILABLE:
        raise ModuleNotFoundError(
            "`jiwer` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )
    cer_value = jiwer.cer(reference, hypothesis)
    if isinstance(cer_value, dict):
        return float(cer_value.get("cer", 0.0))  # Extract CER if returned in a dictionary
    return float(cer_value)  # Ensure output is always a float



"""This module represents the API for the running speech emotion recognition within the senselab package.

Currently, it supports only Hugging Face models, with plans to include more in the future.
Users can specify the audio clips to classify, the classification model, the preferred device,
and the model-specific parameters, and senselab handles the rest.
"""

import warnings
from enum import Enum
from typing import Any, List, Optional

from transformers import AutoConfig

from senselab.audio.data_structures import Audio, AudioClassificationResult
from senselab.audio.tasks.classification.huggingface import HuggingFaceAudioClassifier
from senselab.utils.data_structures import DeviceType, HFModel, SenselabModel, logger


class SERType(Enum):
    """SER types for determining model output behaviors."""

    DISCRETE = "discrete"
    CONTINUOUS = "continuous"
    VALENCE = "valence"
    NOT_RECOGNIZED = "not recognized"


def classify_emotions_from_speech(
    audios: List[Audio],
    model: SenselabModel,
    device: Optional[DeviceType] = None,
    **kwargs: Any,  # noqa: ANN401
) -> List[AudioClassificationResult]:
    """Classify all audios using the given speech emotion recognition model.

    Args:
        audios (List[Audio]): The list of audio objects to be classified.
        model (SenselabModel): The model used for classification, should be trained for recognizing emotions.
        device (Optional[DeviceType]): The device to run the model on (default is None).
        **kwargs: Additional keyword arguments to pass to the classification function.

    Returns:
        List[AudioClassificationResult]: The list of speech emotion recognition results.

    Todo:
        - Include more models (e.g., speechbrain, nvidia)
    """
    if isinstance(model, HFModel):
        model_info = model.get_model_info()
        tags = model_info.tags or []

        ser_type = _get_ser_type(model)

        if (
            not ("speech-emotion-recognition" in tags or "emotion-recognition" in tags)
            and ser_type == SERType.NOT_RECOGNIZED
        ):
            raise ValueError(
                f"The model '{model.path_or_uri}' is not suitable for speech emotion recognition. Please +"
                "validate that it has the correct tags or use the more generic "
                + "'audio_classification_with_hf_models' function."
            )

        if ser_type == SERType.CONTINUOUS:
            output_function_to_apply = kwargs.get("function_to_apply", None)
            if output_function_to_apply:
                if output_function_to_apply != "none":
                    warnings.warn("""Senselab predicts that you are using a continuous SER model but have
                                  specified the parameter `function_to_apply` as something other than none. This
                                  might create side effects when dealing with continuous values that do not
                                  necessarily represent probabilities.""")
            else:
                kwargs["function_to_apply"] = "none"
                logger.info("""Senselab predicts that you are using a continuous SER and have not specified the
                            parameter `function_to_apply`. We are setting this to none such that the model
                            outputs are reported directly rather than being passed through a softmax or sigmoid.""")
        return HuggingFaceAudioClassifier.classify_audios_with_transformers(
            audios=audios, model=model, device=device, **kwargs
        )
    else:
        raise NotImplementedError(
            "Only Hugging Face models are supported for now. We aim to support more models in the future."
        )


def _get_ser_type(model: HFModel) -> SERType:
    """Get the type of SER the model is likely used for based on the labels it is set to predict."""
    config = AutoConfig.from_pretrained(model.path_or_uri)
    id2label = config.id2label
    # print(id2label)
    if id2label:
        # id2label = model.model_info.config['id2label']
        labels = list(id2label.values())
        if "positive" in labels and "negative" in labels and "neutral" in labels:  # Simple, discrete
            return SERType.VALENCE
        if "valence" in labels or "arousal" in labels or "dominance" in labels:  # conitnuous emotions
            return SERType.CONTINUOUS
        if (
            "happy" in labels
            or "happiness" in labels
            or "joy" in labels
            or "sad" in labels
            or "sadness" in labels
            or "fear" in labels
            or "fearful" in labels
            or "anger" in labels
            or "angry" in labels
            or "disgust" in labels
            or "neutral" in labels
            or "calm" in labels
            or "surprised" in labels
        ):
            return SERType.DISCRETE
    return SERType.NOT_RECOGNIZED



"""This module provides utilities for reading and saving (multiple) audio files using Pydra."""

import os
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import pydra

from senselab.audio.data_structures import Audio


def read_audios(
    file_paths: List[str | os.PathLike],
    cache_dir: Optional[Union[str, os.PathLike]] = None,
    plugin: str = "serial",
    plugin_args: Dict[str, Any] = {},
) -> List[Audio]:
    """Read and process a list of audio files using Pydra workflow.

    Args:
        file_paths (List[str]): List of paths to audio files.
        cache_dir (str, optional): Directory for caching intermediate results. Defaults to None.
        plugin (str, optional): Pydra plugin to use for workflow submission. Defaults to "serial".
        plugin_args (dict, optional): Additional arguments for the Pydra plugin. Defaults to {}.

    Returns:
        List[Audio]: A list of Audio objects containing the waveform and sample rate for each processed file.
    """

    @pydra.mark.task
    def load_audio_file(file_path: str | os.PathLike) -> Any:  # noqa: ANN401
        """Load an audio file and return an Audio object.

        Args:
            file_path (str): Path to the audio file.

        Returns:
            Audio: An instance of the Audio class containing the waveform and sample rate.
        """
        return Audio(filepath=file_path)

    # Create the workflow
    wf = pydra.Workflow(name="read_audio_files_workflow", input_spec=["x"], cache_dir=cache_dir)
    wf.split("x", x=file_paths)
    wf.add(load_audio_file(name="load_audio_file", file_path=wf.lzin.x))
    wf.set_output([("processed_files", wf.load_audio_file.lzout.out)])

    # Run the workflow
    with pydra.Submitter(plugin=plugin, **plugin_args) as sub:
        sub(wf)

    # Collect and return the results
    outputs = wf.result()
    return [output.output.processed_files for output in outputs]


def save_audios(
    audio_tuples: Sequence[Tuple[Audio, Union[str, os.PathLike]]],
    save_params: Dict[str, Any] = {},
    cache_dir: Optional[Union[str, os.PathLike]] = None,
    plugin: str = "serial",
    plugin_args: Dict[str, Any] = {},
) -> None:
    """Save a list of Audio objects to specified files using Pydra workflow.

    Args:
        audio_tuples (Sequence[Tuple[Audio, Union[str, os.PathLike]]): Sequence of tuples where each tuple contains
            an Audio object, its output path (str or os.PathLike).
        save_params (dict, optional): Additional parameters for saving audio files.
            Defaults to {}
        cache_dir (str, optional): Directory for caching intermediate results.
            Defaults to None.
        plugin (str, optional): Pydra plugin to use for workflow submission.
            Defaults to "serial".
        plugin_args (dict, optional): Additional arguments for the Pydra plugin.
            Defaults to {}.

    Raises:
        RuntimeError: If any output directory in the provided paths does not exist or is not writable.
    """

    @pydra.mark.task
    def _extract_audio(audio_tuple: Tuple[Audio, Union[str, os.PathLike]]) -> Any:  # noqa: ANN401
        """Extract the Audio object from the tuple."""
        return audio_tuple[0]

    @pydra.mark.task
    def _extract_output_path(audio_tuple: Tuple[Audio, Union[str, os.PathLike]]) -> Union[str, os.PathLike]:
        """Extract the output path from the tuple."""
        return audio_tuple[1]

    @pydra.mark.task
    def _save_audio(audio: Audio, file_path: str, save_params: Dict[str, Any]) -> None:
        """Save an Audio object to a file.

        Args:
            audio (Audio): The Audio object to save.
            file_path (str): Path to save the audio file.
            save_params (dict): Additional parameters for saving audio files.
        """
        audio.save_to_file(file_path, **save_params)

    # Create the workflow
    wf = pydra.Workflow(name="save_audio_files_workflow", input_spec=["audio_tuples"], cache_dir=cache_dir)
    wf.split("audio_tuples", audio_tuples=audio_tuples)
    wf.add(_extract_audio(name="_extract_audio", audio_tuple=wf.lzin.audio_tuples))
    wf.add(_extract_output_path(name="_extract_output_path", audio_tuple=wf.lzin.audio_tuples))
    wf.add(
        _save_audio(
            name="_save_audio",
            audio=wf._extract_audio.lzout.out,
            file_path=wf._extract_output_path.lzout.out,
            save_params=save_params,
        )
    )
    wf.set_output([])

    # Run the workflow
    with pydra.Submitter(plugin=plugin, **plugin_args) as sub:
        sub(wf)



"""This module provides the implementation of language data structures."""

import pycountry
import pycountry.db
from pydantic import BaseModel, field_validator


class Language(BaseModel):
    """Data structure for a language."""

    language_code: str

    @field_validator("language_code", mode="before")
    def validate_language(cls, v: str) -> str:
        """Validate that the language code is valid."""
        lang = (
            pycountry.languages.get(alpha_2=v) or pycountry.languages.get(alpha_3=v) or pycountry.languages.get(name=v)
        )

        if not lang:
            raise ValueError(f"{v} is not a valid ISO language code or name")

        return lang.alpha_3  # ISO 639-3

    @property
    def alpha_2(self) -> str:
        """Get the ISO 639-2 code for the language."""
        return pycountry.languages.get(alpha_3=self.language_code).alpha_2

    @property
    def alpha_3(self) -> str:
        """Get the ISO 639-3 code for the language."""
        return self.language_code

    @property
    def name(self) -> str:
        """Get the name of the language."""
        return pycountry.languages.get(alpha_3=self.language_code).name



"""This module contains the definition of the AudioClassificationResult class."""

from typing import Dict, List, Tuple

from pydantic import BaseModel, model_validator


class AudioClassificationResult(BaseModel):
    """A class to represent a classification result.

    Attributes:
        labels (List[str]): Labels that the classification model is predicting over.
        scores (List[float]): The score for each label the classification model outputs.
    """

    labels: List[str] = []
    scores: List[float] = []

    @model_validator(mode="after")
    def validate_labels_and_scores(self) -> "AudioClassificationResult":
        """Validates that the labels and scores are the same length and sort them."""
        labels = self.labels
        scores = self.scores

        if len(labels) != len(scores):
            raise ValueError("'labels' and 'scores' should be the same length.")

        combined = list(zip(labels, scores))
        combined.sort(key=lambda x: x[1], reverse=True)  # sort scores (and their labels) in reverse order

        sorted_labels, sorted_scores = zip(*combined)
        self.labels = list(sorted_labels)
        self.scores = list(sorted_scores)

        return self

    def get_labels(self) -> List[str]:
        """Get the labels that the model was classifying over.

        Returns:
            List[str]: The full list of labels the model was classifying over,
              are in order of highest to lowest classification score.
        """
        return self.labels

    def get_scores(self) -> List[float]:
        """Get the labels that the model was classifying over.

        Returns:
            List[float]: The full list of scores the model output for each label,
              in order of highest to lowest classification score.
        """
        return self.scores

    def get_label_score_pairs(self) -> List[Tuple[str, float]]:
        """Get the labels and their scores.

        Returns:
            List[Tuple[str,float]]: Pairs of labels and their associated scores from
              the classification model, from highest to lowest score.
        """
        return list(zip(self.labels, self.scores))

    def top_label(self) -> str:
        """Gets the top label based on the highest score."""
        return self.labels[0]

    def top_score(self) -> float:
        """Gets the top score, based on highest value."""
        return self.scores[0]

    def top(self) -> Tuple[str, float]:
        """Gets the top label/score pairing, based on highest score value."""
        return self.labels[0], self.scores[0]

    def top_k_labels(self, k: int = 1) -> List[str]:
        """Gets the top k label(s) based on the highest score."""
        if k < 1 or k > len(self.labels):
            raise ValueError("k needs to be between 1 and the number of labels the model classifies over.")
        return self.labels[0:k]

    def top_k_score(self, k: int = 1) -> List[float]:
        """Gets the top k score(s), based on highest value."""
        if k < 1 or k > len(self.scores):
            raise ValueError("k needs to be between 1 and the number of labels the model classifies over.")
        return self.scores[0:k]

    def top_k(self, k: int = 1) -> List[Tuple[str, float]]:
        """Gets the top k label/score pairing(s), based on highest score value."""
        if k < 1 or k > len(self.labels):
            raise ValueError("k needs to be between 1 and the number of labels the model classifies over.")
        return list(zip(self.labels, self.scores))[0:k]

    @classmethod
    def from_hf_classification_pipeline(cls, results: List[Dict]) -> "AudioClassificationResult":
        """Creates an AudioClassificationResult instance from HuggingFace's pipeline output.

        Args:
            results (List[Dict]): List of dictionaries of the form {"score": score, "label": label}.

        Returns:
            AudioClassificationResult: An instance of AudioClassificationResult.

        Raises:
            ValueError if results dictionaries are not of the form described.
        """
        labels = []
        scores = []

        for classification_result in results:
            if "label" not in classification_result or "score" not in classification_result:
                raise ValueError("Result dictionary is not of the form {'label': label, 'score': score}")
            labels.append(classification_result["label"])
            scores.append(classification_result["score"])

        return cls(labels=labels, scores=scores)



"""This module implements the Pyannote Diarization task."""

import time
from typing import Dict, List, Optional, Union

import torch

from senselab.audio.data_structures import Audio
from senselab.utils.data_structures import DeviceType, PyannoteAudioModel, ScriptLine, _select_device_and_dtype
from senselab.utils.data_structures.logging import logger

try:
    from pyannote.audio import Pipeline
    from pyannote.core import Annotation

    PYANNOTEAUDIO_AVAILABLE = True
except ModuleNotFoundError:
    PYANNOTEAUDIO_AVAILABLE = False


class PyannoteDiarization:
    """A factory for managing Pyannote Diarization pipelines."""

    _pipelines: Dict[str, "Pipeline"] = {}

    @classmethod
    def _get_pyannote_diarization_pipeline(
        cls,
        model: PyannoteAudioModel,
        device: Union[DeviceType, None],
    ) -> "Pipeline":
        """Get or create a Pyannote Diarization pipeline.

        Args:
            model (PyannoteAudioModel): The Pyannote model.
            device (DeviceType): The device to run the model on.

        Returns:
            Pipeline: The diarization pipeline.
        """
        if not PYANNOTEAUDIO_AVAILABLE:
            raise ModuleNotFoundError(
                "`pyannote-audio` is not installed. "
                "Please install senselab audio dependencies using `pip install senselab['audio']`."
            )

        device, _ = _select_device_and_dtype(
            user_preference=device, compatible_devices=[DeviceType.CUDA, DeviceType.CPU]
        )
        key = f"{model.path_or_uri}-{model.revision}-{device}"
        if key not in cls._pipelines:
            pipeline = Pipeline.from_pretrained(checkpoint_path=f"{model.path_or_uri}@{model.revision}").to(
                torch.device(device.value)
            )
            cls._pipelines[key] = pipeline
        return cls._pipelines[key]


def diarize_audios_with_pyannote(
    audios: List[Audio],
    model: Optional[PyannoteAudioModel] = None,
    device: Optional[DeviceType] = None,
    num_speakers: Optional[int] = None,
    min_speakers: Optional[int] = None,
    max_speakers: Optional[int] = None,
) -> List[List[ScriptLine]]:
    """Diarizes a list of audio files using the Pyannote speaker diarization model.

    Args:
        audios (List[Audio]): A list of audio files.
        model (PyannoteAudioModel): The model to use for diarization.
            If None, the default model "pyannote/speaker-diarization-3.1" is used.
        device (Optional[DeviceType]): The device to use for diarization.
        num_speakers (Optional[int]): Number of speakers, when known.
        min_speakers (Optional[int]): Minimum number of speakers. Has no effect when `num_speakers` is provided.
        max_speakers (Optional[int]): Maximum number of speakers. Has no effect when `num_speakers` is provided.

    Returns:
        List[ScriptLine]: A list of ScriptLine objects containing the diarization results.
    """

    def _annotation_to_script_lines(annotation: "Annotation") -> List[ScriptLine]:
        """Convert a Pyannote annotation to a list of script lines.

        Args:
            annotation (Annotation): The Pyannote annotation object.

        Returns:
            List[ScriptLine]: A list of script lines.
        """
        diarization_list: List[ScriptLine] = []
        for segment, _, label in annotation.itertracks(yield_label=True):
            diarization_list.append(ScriptLine(speaker=label, start=segment.start, end=segment.end))
        return diarization_list

    if not PYANNOTEAUDIO_AVAILABLE:
        raise ModuleNotFoundError(
            "`pyannote-audio` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )

    if model is None:
        model = PyannoteAudioModel(path_or_uri="pyannote/speaker-diarization-3.1", revision="main")

    # 16khz comes from the model cards of pyannote/speaker-diarization-3.1
    expected_sample_rate = 16000

    # Check that all audio objects have the correct sampling rate
    for audio in audios:
        if audio.waveform.shape[0] != 1:
            raise ValueError(f"Audio waveform must be mono (1 channel), but got {audio.waveform.shape[0]} channels")
        if audio.sampling_rate != expected_sample_rate:
            raise ValueError(
                "Audio sampling rate "
                + str(audio.sampling_rate)
                + " does not match expected "
                + str(expected_sample_rate)
            )

    # Take the start time of the model initialization
    start_time_model = time.time()
    pipeline = PyannoteDiarization._get_pyannote_diarization_pipeline(model=model, device=device)
    end_time_model = time.time()
    elapsed_time_model = end_time_model - start_time_model
    logger.info(f"Time taken to initialize the pyannote model: {elapsed_time_model:.2f} seconds")

    # Perform diarization
    start_time_diarization = time.time()
    results: List[List[ScriptLine]] = []
    for audio in audios:
        diarization = pipeline(
            {"waveform": audio.waveform, "sample_rate": audio.sampling_rate},
            num_speakers=num_speakers,
            min_speakers=min_speakers,
            max_speakers=max_speakers,
        )
        results.append(_annotation_to_script_lines(diarization))
    end_time_diarization = time.time()
    elapsed_time_diarization = end_time_diarization - start_time_diarization
    logger.info(f"Time taken to perform diarization: {elapsed_time_diarization:.2f} seconds")

    return results



"""This module implements pose estimation functionality."""

import os
from abc import ABC, abstractmethod
from typing import List

import numpy as np

try:
    import cv2

    CV2_AVAILABLE = True
except ModuleNotFoundError:
    CV2_AVAILABLE = False

try:
    import mediapipe as mp
    from mediapipe.tasks import python
    from mediapipe.tasks.python import vision

    MEDIAPIPE_AVAILABLE = True
except ModuleNotFoundError:
    MEDIAPIPE_AVAILABLE = False

try:
    from ultralytics import YOLO

    YOLO_AVAILABLE = True
except ModuleNotFoundError:
    YOLO_AVAILABLE = False

from senselab.video.data_structures.pose import (
    ImagePose,
    IndividualPose,
    MediaPipePoseLandmark,
    PoseModel,
    YOLOPoseLandmark,
)
from senselab.video.tasks.pose_estimation.utils import (
    MEDIAPIPE_KEYPOINT_MAPPING,
    YOLO_KEYPOINT_MAPPING,
    get_model,
)


class PoseEstimator(ABC):
    """Abstract base class for pose estimation models.

    Attributes:
        model_path (str): Path to the PoesEstimator model file.

    Methods:
        estimate(image): Estimate poses in the given image.
        estimate_from_path(image_path): Estimate poses in the image loaded from the specified path.
    """

    model_path: str

    @abstractmethod
    def __init__(self, model_type: str) -> None:
        """Initialize the PoseEstimator.

        Args:
            model_type (str): Type of model to use.
        """
        pass

    @abstractmethod
    def estimate(self, image: np.ndarray) -> ImagePose:
        """Estimate poses in an image.

        Args:
            image (np.ndarray): Input image in RGB format.

        Returns:
            ImagePose: An object containing detected poses and metadata.
        """
        pass

    @abstractmethod
    def estimate_from_path(self, image_path: str) -> ImagePose:
        """Estimate poses in an image loaded from a file path.

        Args:
            image_path (str): Path to the input image file.

        Returns:
            ImagePose: An object containing detected poses and metadata.
        """
        pass


class MediaPipePoseEstimator(PoseEstimator):
    """MediaPipe implementation of pose estimation.

    Attributes:
        model_path (str): Path to the MediaPipe model file.
        num_individuals (int): Number of individuals to detect.
    """

    def __init__(self, model_type: str = "lite") -> None:
        """Initialize the MediaPipePoseEstimator.

        Args:
            model_type (str): Type of MediaPipe model to use ('lite', 'full', 'heavy').
        """
        if not MEDIAPIPE_AVAILABLE:
            raise ModuleNotFoundError(
                "`mediapipe` is not installed. "
                "Please install senselab video dependencies using `pip install senselab['video']`."
            )

        self.model_path = get_model("mediapipe", model_type)
        self.num_individuals = 1
        self._detector = None

    def detector(self, num_individuals: int = 1) -> "vision.PoseLandmarker":
        """Initialization of the MediaPipe detector.

        Args:
            num_individuals (int): Maximum number of individuals to detect. Defaults to 1.

        Returns:
            vision.PoseLandmarker: A MediaPipe pose landmarker object.
        """
        if self._detector is None or self.num_individuals != num_individuals:
            base_options = python.BaseOptions(model_asset_path=self.model_path)
            options = vision.PoseLandmarkerOptions(
                base_options=base_options, output_segmentation_masks=True, num_poses=num_individuals
            )
            self.num_individuals = num_individuals
            self._detector = vision.PoseLandmarker.create_from_options(options)
        return self._detector

    def estimate(self, image: np.ndarray, num_individuals: int = 1) -> ImagePose:
        """Estimate poses using MediaPipe.

        Args:
            image (np.ndarray): Input image in RGB format.
            num_individuals (int): Maximum number of individuals to detect. Defaults to 1.

        Returns:
            ImagePose: An object containing detected poses and metadata.

        Raises:
            ValueError: If `num_individuals` is not a positive integer.
        """
        if not isinstance(num_individuals, int) or num_individuals < 0:
            raise ValueError("Number of individuals must be an integer >=0")

        mp_image = mp.Image(image_format=mp.ImageFormat.SRGB, data=image)

        detection_result = self.detector(num_individuals).detect(mp_image)

        individuals = self.extract_landmarks(detection_result)
        return ImagePose(
            image=image,
            individuals=individuals,
            model=PoseModel.MEDIAPIPE,
        )

    def estimate_from_path(self, image_path: str, num_individuals: int = 1) -> ImagePose:
        """Estimate poses in image from file path using MediaPipe.

        Args:
            image_path (str): Path to the input image file.
            num_individuals (int): Maximum number of individuals to detect. Defaults to 1.

        Returns:
            ImagePose: An object containing detected poses and metadata.

        Raises:
            FileNotFoundError: If the specified image file does not exist.
            ValueError: If `num_individuals` is not a positive integer.
        """
        if not CV2_AVAILABLE:
            raise ModuleNotFoundError(
                "`opencv-python` is not installed. "
                "Please install senselab video dependencies using `pip install senselab['video']`."
            )

        if not os.path.exists(image_path):
            raise FileNotFoundError(f"Image not found at: {image_path}")

        # Load and process the image
        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        return self.estimate(image, num_individuals)

    def extract_landmarks(self, detection_result: "vision.PoseLandmarkerResult") -> List[IndividualPose]:
        """Extract normalized and world landmarks from the detection result.

        Args:
            detection_result (vision.PoseLandmarkerResult): The result from the MediaPipe pose detector.

        Returns:
            List[IndividualPose]: A list of `IndividualPose` objects representing detected individuals.
        """
        individuals = []
        for idx, (norm_landmarks, world_landmarks) in enumerate(
            zip(detection_result.pose_landmarks, detection_result.pose_world_landmarks)
        ):
            norm_dict = {
                MEDIAPIPE_KEYPOINT_MAPPING.get(i, f"keypoint_{i}"): MediaPipePoseLandmark(
                    x=lm.x, y=lm.y, z=lm.z, visibility=lm.visibility
                )
                for i, lm in enumerate(norm_landmarks)
            }

            world_dict = {
                MEDIAPIPE_KEYPOINT_MAPPING.get(i, f"keypoint_{i}"): MediaPipePoseLandmark(
                    x=lm.x, y=lm.y, z=lm.z, visibility=norm_landmarks[i].visibility
                )
                for i, lm in enumerate(world_landmarks)
            }

            individual = IndividualPose(
                individual_index=idx,
                normalized_landmarks=norm_dict,
                world_landmarks=world_dict,
            )
            individuals.append(individual)
        return individuals


class YOLOPoseEstimator(PoseEstimator):
    """YOLO implementation of pose estimation.

    Attributes:
        model_path (str): Path to the YOLO model file.
    """

    def __init__(self, model_type: str = "8n") -> None:
        """Initialize the YOLOPoseEstimator.

        Args:
            model_type (str): Type of YOLO model to use (e.g. '8n', '11p', '11s').
        """
        if not YOLO_AVAILABLE:
            raise ModuleNotFoundError(
                "`ultralytics` is not installed. "
                "Please install senselab video dependencies using `pip install senselab['video']`."
            )

        self.model_path = get_model("yolo", model_type)
        self._model = YOLO(self.model_path)

    def estimate(self, image: np.ndarray) -> ImagePose:
        """Estimate poses using YOLO.

        Args:
            image (np.ndarray): Input image in RGB format.

        Returns:
            ImagePose: An object containing detected poses and metadata.
        """
        results = self._model(image, verbose=False)

        if results[0].keypoints is None or results[0].keypoints.data.numel() == 0:
            # No individuals detected
            return ImagePose(
                image=image,
                individuals=[],
                model=PoseModel.YOLO,
            )

        individuals = []
        for idx, person_keypoints in enumerate(results[0].keypoints):
            confidence_values = person_keypoints.conf.squeeze()  # Extract confidence values

            normalized_dict = {
                YOLO_KEYPOINT_MAPPING.get(i, f"keypoint_{i}"): YOLOPoseLandmark(
                    x=kp[0].item(), y=kp[1].item(), confidence=confidence_values[i].item()
                )
                for i, kp in enumerate(person_keypoints.xyn[0])
            }

            individual = IndividualPose(
                individual_index=idx,
                normalized_landmarks=normalized_dict,
            )
            individuals.append(individual)

        return ImagePose(
            image=image,
            individuals=individuals,
            model=PoseModel.YOLO,
        )

    def estimate_from_path(self, image_path: str) -> ImagePose:
        """Estimate poses in image from file path using YOLO.

        Args:
            image_path (str): Path to the input image file.

        Returns:
            ImagePose: An object containing detected poses and metadata.

        Raises:
            FileNotFoundError: If the specified image file does not exist.
        """
        if not CV2_AVAILABLE:
            raise ModuleNotFoundError(
                "`opencv-python` is not installed. "
                "Please install senselab video dependencies using `pip install senselab['video']`."
            )

        if not os.path.exists(image_path):
            raise FileNotFoundError(f"Image not found at: {image_path}")

        # Load and process the image
        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        return self.estimate(image)



"""Functionality for Pydra."""

from typing import Iterator

import numpy as np
import torch
from pydra.utils.hash import Cache, bytes_repr_sequence_contents, register_serializer

try:
    import opensmile

    OPENSMILE_AVAILABLE = True
except ModuleNotFoundError:
    OPENSMILE_AVAILABLE = False


@register_serializer(torch.Tensor)
def bytes_repr_arraylike(obj: torch.Tensor, cache: Cache) -> Iterator[bytes]:
    """Register a serializer for Torch tensors that allows Pydra to properly use them."""
    yield f"{obj.__class__.__module__}{obj.__class__.__name__}:".encode()
    array = np.asanyarray(obj)
    yield f"{array.size}:".encode()
    if array.dtype == "object":
        yield from bytes_repr_sequence_contents(iter(array.ravel()), cache)
    else:
        yield array.tobytes(order="C")


if OPENSMILE_AVAILABLE:

    @register_serializer(opensmile.Smile)
    def bytes_repr_smile(obj: opensmile.Smile, _cache: Cache) -> Iterator[bytes]:
        """Serializer for opensmile.Smile.

        This function registers a custom serializer for instances of `opensmile.Smile`,
        allowing Pydra's caching system to recognize and hash these objects based on
        their configurations. By encoding essential attributes to bytes, we ensure that
        identical configurations produce the same hash, facilitating efficient workflow caching.

        Key Attributes Serialized:
            - `feature_set`: The OpenSMILE feature set, e.g., `eGeMAPSv02`.
            - `feature_level`: The feature level, e.g., `Functionals` or `LowLevelDescriptors`.
            - `options`: A dictionary containing additional configurations for feature extraction.
            - `logfile`: The log file path, if logging is enabled.
            - `verbose`: Boolean indicating verbosity in logging.
            - `column_names`: Column names of features generated by OpenSMILE, represented as an index.
            - `feature_names`: List of specific feature names extracted by OpenSMILE.
            - `hop_dur`: The hop duration for windowed feature extraction, if applicable.
            - `name`: Name identifier for the OpenSMILE instance.
            - `num_channels`: Number of audio channels expected by the instance.
            - `num_features`: Number of features generated for each frame.
            - `params`: Dictionary of internal configuration parameters such as `sampling_rate`, `channels`,
            `mixdown`, `resample`, and other settings impacting feature extraction.
            - `process_func_applies_sliding_window`: Indicates if a sliding window is applied in feature extraction.
            - `win_dur`: Duration of each window frame, if applicable.

        Args:
            obj (opensmile.Smile): The `opensmile.Smile` instance to be serialized.
            _cache (Cache): The Pydra cache object.

        Usage:
            This serializer is automatically used by Pydra to calculate a unique hash for `opensmile.Smile`
            objects in workflows, ensuring consistent hashing based on the object's configurations. The
            serializer helps avoid hash collisions in cases where `opensmile.Smile` instances have the same
            internal settings but different object IDs in memory.

        Returns:
            Iterator[bytes]: Byte-encoded representations of each serialized attribute.
        """
        _ = _cache  # This is just to silence the unused parameter warning

        yield f"{obj.__class__.__module__}{obj.__class__.__name__}:".encode()

        # Serialize key configuration attributes
        yield f"feature_set:{obj.feature_set}".encode()
        yield f"feature_level:{obj.feature_level}".encode()
        yield f"options:{obj.options}".encode()
        yield f"logfile:{obj.logfile}".encode()
        yield f"verbose:{obj.verbose}".encode()
        yield f"column_names:{obj.column_names}".encode()
        yield f"feature_names:{obj.feature_names}".encode()
        yield f"hop_dur:{obj.hop_dur}".encode()
        yield f"name:{obj.name}".encode()
        yield f"num_channels:{obj.num_channels}".encode()
        yield f"num_features:{obj.num_features}".encode()
        yield f"params:{obj.params}".encode()
        yield f"process_func_applies_sliding_window:{obj.process_func_applies_sliding_window}".encode()
        yield f"win_dur:{obj.win_dur}".encode()


# TODO: Ignore this for now but need to decide how to incorporate Pydra into the package
# Pydra runner
# need function that allows for marking a task (could be obfuscated internally)
# need function that runs the actual code/does the parallelization
# getting results is a bit clunky right now so could clean that up
# internal Pydra wouldn't need obfuscations for add/split but external API might be useful
# perhaps dictionary structure could be intuitive? linked list?

# what if we had a simple split/run structure:
# user: run X task with Y audios (Satra example); perhaps extraneous information of DeviceType, Model to use
#   - kinda easy example of just create the pydra setup ourselves, no user worries
# Senselab "pipelines"/Tasks:

# example:
# RAVDESS audios:
# - need to resample
# - perhaps run loudness normalization
# - run SER
# - run evaluation

# get RAVDESS audios
# split into batches
#   - run batches through resmapling
#   - pass to normalization
#   - run SER
#   - get partial evaluation
# - combine evaluation
# - give average



"""This module implements plotting methods for utilities."""

from typing import List

import matplotlib.cm as cm
import matplotlib.pyplot as plt

from senselab.utils.data_structures import ScriptLine


def plot_transcript(transcript: ScriptLine) -> None:
    """Plots the transcript visualization over time.

    Args:
        transcript (ScriptLine): The transcript object containing chunks of text with start and end times.

    Returns:
        None

    Todo:
        - Add option to save the plot
        - Add option to choose the size of the Figure
        - Add check if transcript contains chunks with time information
    """
    if transcript.chunks is None:
        raise ValueError("The transcript does not contain any chunks.")

    chunks = transcript.chunks

    texts = [chunk.text for chunk in chunks]
    start_times = []
    end_times = []
    for chunk in chunks:
        # Ensure that chunks have start and end times
        if chunk.start is None or chunk.end is None:
            raise ValueError("Each chunk must have both start and end times.")
        else:
            start_times.append(chunk.start)
            end_times.append(chunk.end)

    # Create a figure and axis
    _, ax = plt.subplots(figsize=(12, 6))

    # Plot each text segment and add text label
    for i, text in enumerate(texts):
        if start_times[i] is not None and end_times[i] is not None:
            ax.plot([start_times[i], end_times[i]], [i, i], marker="o")
            if text:
                ax.text((start_times[i] + end_times[i]) / 2, i, text, ha="center", va="bottom")

    # Setting labels and title
    ax.set_yticks(range(len(texts)))
    ax.set_yticklabels([])
    ax.set_xlabel("Time (seconds)")
    ax.set_title("Transcript Visualization Over Time")

    # Show the plot
    plt.show()


def plot_segment(segments: List[ScriptLine]) -> None:
    """Plots the segments of the transcript over time.

    Args:
        segments (List[ScriptLine]): The segments object containing segments with start and end times and a label.

    Returns:
        None

    Todo:
        - Add option to save the plot
        - Add option to choose the size of the Figure
        - Add check if transcript contains segments with time information
    """
    start_times = []
    end_times = []
    labels = []

    for segment in segments:
        # Ensure that segments have start and end times and a label
        if segment.start is None or segment.end is None or segment.speaker is None:
            raise ValueError("Each segment must have start and end times and a label.")
        else:
            start_times.append(segment.start)
            end_times.append(segment.end)
            labels.append(segment.speaker)

    # Create a figure and axis
    _, ax = plt.subplots(figsize=(12, 6))

    # Create a color map based on unique labels
    unique_labels = list(set(labels))
    color_map = cm.get_cmap("tab10", len(unique_labels))  # 'tab10' provides 10 distinct colors
    label_to_color = {label: color_map(i) for i, label in enumerate(unique_labels)}
    label_to_y_value = {label: i for i, label in enumerate(unique_labels)}  # Assign y-value based on label index

    # Plot each segment and add text label with color
    for i, label in enumerate(labels):
        color = label_to_color[label]
        y_value = label_to_y_value[label]  # Get y-value based on label
        ax.plot([start_times[i], end_times[i]], [y_value, y_value], marker="o", color=color, linewidth=2)
        ax.text((start_times[i] + end_times[i]) / 2, y_value, label, ha="center", va="bottom", color=color)

    # Setting labels and title
    ax.set_yticks(range(len(unique_labels)))
    ax.set_yticklabels(unique_labels)
    ax.set_xlabel("Time (seconds)")
    ax.set_title("Segment Visualization Over Time")

    # Show the plot
    plt.show()



"""This module provides a wrapper method for extracting embeddings from text using different models."""

from typing import List, Optional

import torch

from senselab.text.tasks.embeddings_extraction.huggingface import HFFactory
from senselab.text.tasks.embeddings_extraction.sentence_transformers import SentenceTransformerFactory
from senselab.utils.data_structures import DeviceType, HFModel, SenselabModel, SentenceTransformersModel


def extract_embeddings_from_text(
    pieces_of_text: List[str], model: SenselabModel, device: Optional[DeviceType] = None
) -> List[torch.Tensor]:
    """Extracts embeddings from a list of strings using the specified model.

    Args:
        pieces_of_text (List[str]): A list of strings to extract embeddings from.
        model (SenselabModel): The model used for extracting embeddings.
        device (Optional[DeviceType]): The device to run the model on. Defaults to None.

    Returns:
        List[torch.Tensor]: A list of embeddings for the input strings.
    """
    if isinstance(model, SentenceTransformersModel):
        return SentenceTransformerFactory.extract_text_embeddings(pieces_of_text, model, device)
    elif isinstance(model, HFModel):
        return HFFactory.extract_text_embeddings(pieces_of_text, model, device)
    raise ValueError("Unsupported model type for text embedding extraction.")



"""This module provides the implementation of file utilities."""

import os
from pathlib import Path
from typing import List

from pydantic import BaseModel, FilePath, field_validator

from senselab.utils.config import get_config


class File(BaseModel):
    """This class represents a file."""

    filepath: FilePath

    @property
    def type(self) -> str:
        """Determine the type of the file based on its extension."""
        extension = os.path.splitext(self.filepath)[1].lower()
        for type_name, type_info in get_config()["files"].items():
            if extension in type_info["extensions"]:
                return type_name
        raise ValueError("File type could not be determined from extension.")

    @field_validator("filepath")
    def validate_filepath(cls, v: FilePath) -> FilePath:
        """Validate the filepath."""
        if not os.path.exists(v):
            raise ValueError(f"File {v} does not exist")

        file_extension = os.path.splitext(v)[1].lower()
        valid_extensions = [ext for types in get_config()["files"].values() for ext in types["extensions"]]

        if file_extension not in valid_extensions:
            raise ValueError(
                f"Unsupported file extension: {file_extension}." f"Supported extensions: {valid_extensions}"
            )

        return v


def from_strings_to_files(list_of_files: List[str]) -> List[File]:
    """Create a list of `File` objects from a list of strings."""
    return [File(filepath=Path(file)) for file in list_of_files]


def get_common_directory(files: List[str]) -> str:
    """A function to get the common directory from a list of file paths.

    Parameters:
    - files: a list of file paths

    Returns:
    - the common directory among the file paths
    """
    if len(files) == 1:
        # Ensure the single path's directory ends with a separator
        common_path = os.path.dirname(files[0])
    else:
        # Use commonpath to find the common directory for multiple files
        common_path = os.path.commonpath(files)

    # Check if the path ends with the os separator, add if not
    if not common_path.endswith(os.sep):
        common_path += os.sep

    return common_path



"""This module provides the API for the senselab speech enhancement task."""

from typing import List, Optional

from senselab.audio.data_structures import Audio
from senselab.audio.tasks.speech_enhancement.speechbrain import SpeechBrainEnhancer
from senselab.utils.data_structures import DeviceType, SpeechBrainModel


def enhance_audios(
    audios: List[Audio],
    model: Optional[SpeechBrainModel] = None,
    device: Optional[DeviceType] = None,
) -> List[Audio]:
    """Enhances all audios using the given model.

    Args:
        audios (List[Audio]): The list of audio objects to be enhanced.
        model (SenselabModel): The model used for enhancement.
            If None, the default model "speechbrain/sepformer-wham16k-enhancement" is used.
        device (Optional[DeviceType]): The device to run the model on (default is None).

    Returns:
        List[Audio]: The list of enhanced audio objects.
    """
    if model is None:
        model = SpeechBrainModel(path_or_uri="speechbrain/sepformer-wham16k-enhancement", revision="main")

    if isinstance(model, SpeechBrainModel):
        return SpeechBrainEnhancer.enhance_audios_with_speechbrain(audios=audios, model=model, device=device)
    else:
        raise NotImplementedError(
            "Only SpeechBrain models are supported for now. We aim to support more models in the future."
        )



"""This module implements some utilities to extract embeddings from self-supervised models."""

import os
from typing import List, Optional, Union

import torch

from senselab.audio.data_structures import Audio
from senselab.audio.tasks.ssl_embeddings.self_supervised_features import SSLEmbeddingsFactory
from senselab.utils.data_structures import DeviceType, HFModel, SenselabModel


def extract_ssl_embeddings_from_audios(
    audios: List[Audio],
    model: SenselabModel,
    cache_dir: Optional[Union[str, os.PathLike]] = None,
    device: Optional[DeviceType] = None,
) -> List[torch.Tensor]:
    """Extract embedding of audio signals from pre-trained SSL models.

    Args:
        audios (List[Audio]): A list of Audio objects containing the audio signals and their properties.
        model (SenselabModel): The model used to extract their embeddings.
        cache_dir (Optional[str, os.PathLike]): The path to where the model's weights will be saved.
        device (Optional[DeviceType]): The device to run the model on (default is None).

    Returns:
        List[torch.Tensor]: A list of 1d tensors containing the ssl embeddings for each audio file.

    Raises:
        NotImplementedError: If the model is not a Hugging Face model.

    Examples:
        >>> audios = [Audio(filepath="sample.wav")]
        >>> model = HFModel(path_or_uri="facebook/wav2vec2-base", revision="main")
        >>> embeddings = extract_ssl_embeddings_from_audios(audios, model, cache_dir="./", device=DeviceType.CUDA)
        >>> print(embeddings[0].shape)
        [13, 209, 768] ([# of Layers, Time Frames, Embedding Size])

    Todo:
        - Make the API compatible with other models than Hugging Face.
    """
    if isinstance(model, HFModel):
        return SSLEmbeddingsFactory.extract_ssl_embeddings(
            audios=audios, model=model, cache_dir=cache_dir, device=device
        )
    else:
        raise NotImplementedError("The specified model is not supported for now.")



"""This module provides a factory for managing Hugging Face audio classification pipelines.

To ensure correct functionality, call `classify_audios_with_transformers` serially or,
if you need to process multiple audios in parallel, pass the entire list of audios to the
function at once, rather than calling the function with one audio at a time.
"""

import time
from typing import Dict, List, Literal, Optional

from transformers import pipeline

from senselab.audio.data_structures import Audio, AudioClassificationResult
from senselab.utils.data_structures import DeviceType, HFModel, _select_device_and_dtype
from senselab.utils.data_structures.logging import logger


class HuggingFaceAudioClassifier:
    """A factory for managing Hugging Face audio classification pipelines."""

    _pipelines: Dict[str, pipeline] = {}

    @classmethod
    def _get_hf_audio_classification_pipeline(
        cls,
        model: HFModel,
        top_k: Optional[int],
        function_to_apply: Literal["softmax", "sigmoid", "none"],
        batch_size: int,
        device: Optional[DeviceType] = None,
    ) -> pipeline:
        """Get or create a Hugging Face audio classification pipeline.

        Args:
            model (HFModel): The Hugging Face model.
            top_k (int): Number of top labels that will be returned by the model.
            function_to_apply (str): Function to apply to the model's output, defaults to softmax.
            batch_size (int): The batch size for processing.
            device (Optional[DeviceType]): The device to run the model on.

        Returns:
            pipeline: The Audio Classification pipeline.
        """
        device, _ = _select_device_and_dtype(
            user_preference=device, compatible_devices=[DeviceType.CUDA, DeviceType.CPU]
        )
        key = f"{model.path_or_uri}-{model.revision}-{top_k}-" f"{function_to_apply}-{batch_size}-{device.value}"
        if key not in cls._pipelines:
            cls._pipelines[key] = pipeline(
                "audio-classification",
                model=model.path_or_uri,
                revision=model.revision,
                # top_k=top_k, #TODO: this causes a bug in the pipeline that has been reported to transformers
                # https://github.com/huggingface/transformers/issues/35736
                function_to_apply=function_to_apply,  # TODO: parameter ignored in transformer code, bug reported
                # https://github.com/huggingface/transformers/issues/35739
                device=device.value,
            )
        return cls._pipelines[key]

    @classmethod
    def classify_audios_with_transformers(
        cls,
        audios: List[Audio],
        model: HFModel,
        top_k: Optional[int] = None,
        function_to_apply: Literal["softmax", "sigmoid", "none"] = "softmax",
        batch_size: int = 16,
        device: Optional[DeviceType] = None,
    ) -> List[AudioClassificationResult]:
        """Transcribes all audio samples in the dataset.

        Args:
            audios (List[Audio]): The list of audio objects to be classified.
            model (HFModel): The Hugging Face model used for classification.
                There is no default model since this task covers a wide range of models.
            top_k (Optional[int]): the number of top labels for the model to return.
                If no top_k, the model returns all of the labels in the configuration file.
            function_to_apply (str): The function to apply to the model's outputs (default is softmax).
            batch_size (int): The batch size for processing (default is 16).
            device (Optional[DeviceType]): The device to run the model on (default is None).

        Returns:
            List[AudioClassificationResult]: The list of classification results,
                where each result contains the classified label and their score.
        """

        def _audio_to_huggingface_dict(audio: Audio) -> Dict:
            """Convert an Audio object to a dictionary that can be used by the transformers pipeline.

            Args:
                audio (Audio): The audio object.

            Returns:
                Dict: The dictionary representation of the audio object.
            """
            return {
                "array": audio.waveform.squeeze().numpy(),
                "sampling_rate": audio.sampling_rate,
            }

        # Take the start time of the pipeline initialization
        start_time_pipeline = time.time()
        # Get the Hugging Face pipeline
        pipe = HuggingFaceAudioClassifier._get_hf_audio_classification_pipeline(
            model=model,
            top_k=top_k,
            function_to_apply=function_to_apply,
            batch_size=batch_size,
            device=device,
        )

        # Take the end time of the pipeline initialization
        end_time_pipeline = time.time()
        # Print the time taken for initialize the hugging face ASR pipeline
        elapsed_time_pipeline = end_time_pipeline - start_time_pipeline
        logger.info(
            f"Time taken to initialize the hugging face audio classification \
            pipeline: {elapsed_time_pipeline:.2f} seconds"
        )

        # Retrieve the expected sampling rate from the Hugging Face model
        expected_sampling_rate = pipe.feature_extractor.sampling_rate

        # Check that all audio objects are mono
        for audio in audios:
            if audio.waveform.shape[0] != 1:
                raise ValueError(f"Stereo audio is not supported. Got {audio.waveform.shape[0]} channels")
            if audio.sampling_rate != expected_sampling_rate:
                raise ValueError(
                    f"Incorrect sampling rate. Expected {expected_sampling_rate}" f", got {audio.sampling_rate}"
                )

        # Convert the audio objects to dictionaries that can be used by the pipeline
        formatted_audios = [_audio_to_huggingface_dict(audio) for audio in audios]

        # Take the start time of the transcription
        start_time_transcription = time.time()
        # Run the pipeline
        classifications = pipe(formatted_audios, top_k=top_k, function_to_apply=function_to_apply)

        # Take the end time of the classification
        end_time_transcription = time.time()
        # Print the time taken for classifying the audios
        elapsed_time_transcription = end_time_transcription - start_time_transcription
        logger.info(f"Time taken for classifying the audios: {elapsed_time_transcription:.2f} seconds")

        # Convert the pipeline output to AudioClassificationResult objects
        return [AudioClassificationResult.from_hf_classification_pipeline(t) for t in classifications]



"""This module contains functions for voice cloning using Coqui TTS."""

from typing import Any, Dict, List, Optional, Union

try:
    from TTS.api import TTS

    TTS_AVAILABLE = True
except ModuleNotFoundError:
    TTS_AVAILABLE = False

from pathlib import Path

from senselab.audio.data_structures import Audio
from senselab.utils.data_structures import CoquiTTSModel, DeviceType, _select_device_and_dtype


class CoquiVoiceCloner:
    """A factory for managing voice cloning pipelines using Coqui TTS."""

    _models: Dict[str, "TTS"] = {}

    @classmethod
    def _get_tts_model(cls, model_id: Union[str, Path], user_preference: Optional[DeviceType] = None) -> "TTS":
        """Get or create a TTS voice conversion pipeline."""
        if not TTS_AVAILABLE:
            raise ModuleNotFoundError(
                "`coqui-tts` is not available. "
                "Please install senselab audio dependencies using `pip install senselab['audio']`."
            )
        device, _ = _select_device_and_dtype(
            user_preference=user_preference, compatible_devices=[DeviceType.CUDA, DeviceType.CPU]
        )
        key = f"{model_id}-{device.value}"
        if key not in cls._models:
            tts = TTS(model_id).to(device=device.value)
            cls._models[key] = tts
        return cls._models[key]

    @classmethod
    def clone_voices(
        cls,
        source_audios: List[Audio],
        target_audios: List[Audio],
        model: Optional[CoquiTTSModel] = None,
        device: Optional[DeviceType] = None,
    ) -> List[Audio]:
        """Clone voices from source audios to target audios using Coqui TTS.

        Args:
            source_audios (List[Audio]): List of source audio objects.
            target_audios (List[Audio]): List of target audio objects.
            model (CoquiTTSModel, optional): The Coqui TTS model to use for voice conversion.
                All Coqui TTS models are supported, including
                "voice_conversion_models/multilingual/multi-dataset/knnvc",
                "voice_conversion_models/multilingual/vctk/freevc24",
                "voice_conversion_models/multilingual/multi-dataset/openvoice_v1",
                "voice_conversion_models/multilingual/multi-dataset/openvoice_v2".
                If None, the default model "voice_conversion_models/multilingual/multi-dataset/knnvc" is used.
            device (Optional[DeviceType], optional): The device to run the model on.
                Defaults to None.
        """
        # Validate model
        if model is None:
            model = CoquiTTSModel(path_or_uri="voice_conversion_models/multilingual/multi-dataset/knnvc")
        if model._scope != "voice_conversion_models":
            all_models = TTS.list_models()
            voice_conversion_models = [m for m in all_models if m.startswith("voice_conversion_models")]
            raise ValueError(
                f"Model {model.path_or_uri} is not a voice conversion model. "
                f"Available voice conversion models: {voice_conversion_models}"
            )

        # Get TTS pipeline
        tts = cls._get_tts_model(model_id=model.path_or_uri, user_preference=device)

        # Validate sampling rates
        audio_config = tts.voice_converter.vc_config.audio

        expected_sample_rate = (
            audio_config.input_sample_rate
            if hasattr(audio_config, "input_sample_rate")
            else getattr(audio_config, "sample_rate", None)
        )

        output_sample_rate = (
            audio_config.output_sample_rate
            if hasattr(audio_config, "output_sample_rate")
            else getattr(audio_config, "sample_rate", None)
        )

        if len(source_audios) != len(target_audios):
            raise ValueError("Number of source and target audios must be the same.")

        for i, (source_audio, target_audio) in enumerate(zip(source_audios, target_audios)):
            source_path = (
                source_audio.filepath() if hasattr(source_audio, "filepath") and source_audio.filepath() else ""
            )
            target_path = (
                target_audio.filepath() if hasattr(target_audio, "filepath") and target_audio.filepath() else ""
            )

            source_desc = f"{source_audio.generate_id()}{f' ({source_path})' if source_path else ''}"
            target_desc = f"{target_audio.generate_id()}{f' ({target_path})' if target_path else ''}"

            if source_audio.waveform.squeeze().dim() != 1 or target_audio.waveform.squeeze().dim() != 1:
                raise ValueError(
                    f"[Pair index {i}] Only mono audio files are supported. "
                    f"Source: {source_desc}, Target: {target_desc}."
                )

            if source_audio.sampling_rate != expected_sample_rate or target_audio.sampling_rate != expected_sample_rate:
                raise ValueError(
                    f"[Pair index {i}] Expected sample rate {expected_sample_rate}, but got "
                    f"{source_audio.sampling_rate} (source) and {target_audio.sampling_rate} (target). "
                    f"Source: {source_desc}, Target: {target_desc}."
                )

            # Ensure mono audio
            source_waveform = source_audio.waveform.squeeze()
            target_waveform = target_audio.waveform.squeeze()
            if source_waveform.dim() > 1 or target_waveform.dim() > 1:
                raise ValueError("Both source and target audios must be mono.")

        cloned_audios = []
        for source_audio, target_audio in zip(source_audios, target_audios):
            # Perform voice conversion
            converted_waveform = tts.voice_conversion(source_wav=source_waveform, target_wav=target_waveform)

            cloned_audios.append(Audio(waveform=converted_waveform, sampling_rate=output_sample_rate))

        return cloned_audios



"""This module contains data structures for audio processing."""

from .audio import Audio, batch_audios, unbatch_audios  # noqa: F401
from .audio_classification_result import AudioClassificationResult  # noqa: F401



"""Data structures relevant for pose estimation."""

from abc import ABC, abstractmethod
from enum import Enum
from typing import Dict, List, Optional

import numpy as np
from pydantic import BaseModel, ConfigDict, field_validator


class PoseModel(str, Enum):
    """Enum representing different pose estimation models.

    Attributes:
        MEDIAPIPE (str): Enum value for MediaPipe pose estimation.
        YOLO (str): Enum value for YOLO pose estimation.
    """

    MEDIAPIPE = "mp"
    YOLO = "yolo"


class PoseLandmark(ABC, BaseModel):
    """Abstract base class for representing pose landmarks.

    Methods:
        as_list() -> List[float]:
            Convert the landmark data to a list of coordinates.
    """

    @abstractmethod
    def as_list(self) -> List[float]:
        """Convert the landmark data to a list of floating-point values.

        Returns:
            List[float]: List of landmark data as floats.
        """
        pass


class MediaPipePoseLandmark(PoseLandmark):
    """Represents a pose landmark detected by MediaPipe.

    Attributes:
        x (float): X-coordinate of the landmark.
        y (float): Y-coordinate of the landmark.
        z (float): Z-coordinate of the landmark.
        visibility (float): Probability of the landmark being visible [0, 1].
    """

    x: float
    y: float
    z: float
    visibility: float

    def as_list(self) -> List[float]:
        """Convert the landmark data to a list.

        Returns:
            List[float]: [x, y, z, visibility] values.
        """
        return [self.x, self.y, self.z, self.visibility]


class YOLOPoseLandmark(PoseLandmark):
    """Represents a pose keypoint detected by YOLO.

    Attributes:
        x (float): X-coordinate of the keypoint.
        y (float): Y-coordinate of the keypoint.
        confidence (float): Confidence score of the detected keypoint [0, 1].
    """

    x: float
    y: float
    confidence: float

    def as_list(self) -> List[float]:
        """Convert the keypoint data to a list.

        Returns:
            List[float]: [x, y, confidence] values.
        """
        return [self.x, self.y, self.confidence]


class IndividualPose(BaseModel):
    """Data structure for the estimated pose of a single individual in an image.

    Attributes:
        individual_index (int): Index of the individual in the detection result.
        normalized_landmarks (Dict[str, PoseLandmark]): Dictionary of body landmarks with normalized coordinates.
        world_landmarks (Optional[Dict[str, PoseLandmark]]): Dictionary of body landmarks with real-world coordinates.
    """

    individual_index: int
    normalized_landmarks: Dict[str, PoseLandmark]
    world_landmarks: Optional[Dict[str, PoseLandmark]] = None

    def get_landmark_coordinates(self, landmark: str, world: bool = False) -> PoseLandmark:
        """Retrieve coordinates for a specific landmark.

        Args:
            landmark (str): Name of the landmark (e.g., "right_ankle").
            world (bool): Whether to retrieve world coordinates. Defaults to False.

        Returns:
            PoseLandmark: Object containing information on the specified landmark.

        Raises:
            ValueError: If the specified landmark is not found.
        """
        landmarks = self.world_landmarks if world else self.normalized_landmarks
        if not landmarks:
            raise ValueError("No landmarks available.")
        if landmark not in landmarks:
            raise ValueError(f"Landmark '{landmark}' not found. Available landmarks: {sorted(landmarks.keys())}")
        return landmarks[landmark]


class ImagePose(BaseModel):
    """Data structure for storing estimated poses of multiple individuals in an image.

    Attributes:
        model (PoseModel): The model used for pose estimation.
        image (np.ndarray): Original image as a NumPy array.
        individuals (List[IndividualPose]): List of IndividualPose objects for each detected individual.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    model: PoseModel
    image: np.ndarray
    individuals: List[IndividualPose]

    @field_validator("image", mode="before")
    def validate_image(cls, value: np.ndarray) -> np.ndarray:
        """Ensures image is a 3D numpy array with three color channels."""
        if not isinstance(value, np.ndarray):
            raise TypeError("Field 'image' must be a NumPy array.")
        if value.ndim != 3 or value.shape[2] != 3:
            raise ValueError("Field 'image' must be a 3D array with three color channels (RGB).")
        return value

    def get_individual(self, individual_index: int) -> IndividualPose:
        """Retrieve a specific individual's pose data.

        Args:
            individual_index (int): Index of the individual to retrieve.

        Returns:
            IndividualPose: Pose data for the specified individual.

        Raises:
            ValueError: If the index is invalid or no individuals are detected.
        """
        if individual_index >= len(self.individuals) or individual_index < 0:
            raise ValueError(
                f"Individual index {individual_index} is invalid. {len(self.individuals)} poses were estimated. "
                f"Valid indices are {f'0 to {len(self.individuals)-1}' if len(self.individuals) > 0 else 'none'}"
            )
        return self.individuals[individual_index]



"""This module contains functions for extracting openSMILE features.

It includes a factory class for managing openSMILE feature extractors, ensuring
each extractor is created only once per feature set and feature level. The main
function, `extract_opensmile_features_from_audios`, applies feature extraction
across a list of audio samples using openSMILE, managed as a Pydra workflow
for parallel processing. This approach supports efficient and scalable feature
extraction across multiple audio files.
"""

try:
    import opensmile

    OPENSMILE_AVAILABLE = True
except ModuleNotFoundError:
    OPENSMILE_AVAILABLE = False

import os
from typing import Any, Dict, List, Optional

import numpy as np
import pydra

from senselab.audio.data_structures import Audio


class OpenSmileFeatureExtractorFactory:
    """A factory for managing openSMILE feature extractors.

    This class creates and caches openSMILE feature extractors, allowing for
    efficient reuse. It ensures only one instance of each feature extractor
    exists per unique combination of `feature_set` and `feature_level`.
    """

    _extractors: Dict[str, "opensmile.Smile"] = {}  # Cache for feature extractors

    @classmethod
    def get_opensmile_extractor(cls, feature_set: str, feature_level: str) -> "opensmile.Smile":
        """Get or create an openSMILE feature extractor.

        Args:
            feature_set (str): The openSMILE feature set.
            feature_level (str): The openSMILE feature level.

        Returns:
            opensmile.Smile: The openSMILE feature extractor.
        """
        if not OPENSMILE_AVAILABLE:
            raise ModuleNotFoundError(
                "`opensmile` is not installed. "
                "Please install senselab audio dependencies using `pip install senselab['audio']`"
            )

        key = f"{feature_set}-{feature_level}"  # Unique key for each feature extractor
        if key not in cls._extractors:  # Check if extractor exists in cache
            # Create and store a new extractor if not found in cache
            cls._extractors[key] = opensmile.Smile(
                feature_set=opensmile.FeatureSet[feature_set],
                feature_level=opensmile.FeatureLevel[feature_level],
            )
        return cls._extractors[key]  # Return cached or newly created extractor


def extract_opensmile_features_from_audios(
    audios: List[Audio],
    feature_set: str = "eGeMAPSv02",
    feature_level: str = "Functionals",
    plugin: str = "serial",
    plugin_args: Optional[Dict[str, Any]] = {},
    cache_dir: Optional[str | os.PathLike] = None,
) -> List[Dict[str, Any]]:
    """Extract openSMILE features from a list of audio files using Pydra workflow.

    This function sets up a Pydra workflow for parallel processing of openSMILE
    feature extraction on a list of audio samples. Each sample's features are
    extracted and formatted as dictionaries.

    Args:
        audios (List[Audio]): The list of audio objects to extract features from.
        feature_set (str): The openSMILE feature set (default is "eGeMAPSv02").
        feature_level (str): The openSMILE feature level (default is "Functionals").
        plugin (str): The Pydra plugin to use (default is "serial").
        plugin_args (Optional[Dict[str, Any]]): Additional arguments for the Pydra plugin.
        cache_dir (Optional[str | os.PathLike]): The path to the Pydra cache directory.

    Returns:
        List[Dict[str, Any]]: A list of dictionaries, each containing extracted features.
    """
    if not OPENSMILE_AVAILABLE:
        raise ModuleNotFoundError(
            "`opensmile` is not installed. Please install the necessary dependencies using:\n"
            "`pip install senselab['audio']`"
        )

    def _extract_feats_from_audio(sample: Audio, smile: opensmile.Smile) -> Dict[str, Any]:
        """Extract features from a single audio sample using openSMILE.

        Args:
            sample (Audio): The audio object.
            smile (opensmile.Smile): The openSMILE feature extractor.

        Returns:
            Dict[str, Any]: The extracted features as a dictionary.
        """
        # Convert audio tensor to a NumPy array for processing
        audio_array = sample.waveform.squeeze().numpy()
        sampling_rate = sample.sampling_rate  # Get sampling rate from Audio object
        try:
            # Process the audio and extract features
            sample_features = smile.process_signal(audio_array, sampling_rate)
            # Convert features to a dictionary and handle single-item lists
            return {
                k: v[0] if isinstance(v, list) and len(v) == 1 else v
                for k, v in sample_features.to_dict("list").items()
            }
        except Exception as e:
            # Log error and return NaNs if feature extraction fails
            filepath = sample.filepath() if hasattr(sample, "filepath") and sample.filepath() else ""
            desc = f"{sample.generate_id()}{f' ({filepath})' if filepath else ''}"
            print(f"Error processing sample {desc}: {e}")
            return {feature: np.nan for feature in smile.feature_names}

    # Decorate the feature extraction function for Pydra
    _extract_feats_from_audio_pt = pydra.mark.task(_extract_feats_from_audio)

    # Obtain the feature extractor using the factory
    smile = OpenSmileFeatureExtractorFactory.get_opensmile_extractor(feature_set, feature_level)

    # Create a Pydra workflow, split it over the list of audio samples
    wf = pydra.Workflow(name="wf", input_spec=["x"], cache_dir=cache_dir)
    wf.split("x", x=audios)  # Each audio is treated as a separate task
    # Add feature extraction task to the workflow
    wf.add(_extract_feats_from_audio_pt(name="_extract_feats_from_audio_pt", sample=wf.lzin.x, smile=smile))

    # Set workflow output to the results of each audio feature extraction
    wf.set_output([("opensmile", wf._extract_feats_from_audio_pt.lzout.out)])

    # Run the workflow using the specified Pydra plugin and arguments
    with pydra.Submitter(plugin=plugin, **plugin_args) as sub:
        sub(wf)

    # Retrieve results from the completed workflow
    outputs = wf.result()

    # Format the outputs into a list of dictionaries
    formatted_output: List[Dict[str, Any]] = []
    for output in outputs:
        # Extract features and organize into a dictionary
        formatted_output_item = {
            f"{feature}": output.output.opensmile[f"{feature}"] for feature in output.output.opensmile
        }
        formatted_output.append(formatted_output_item)  # Append to final output list
    return formatted_output  # Return the list of formatted feature dictionaries



"""This module provides the implementation of config utilities."""

import json
import os
from typing import Any, Dict


def get_config() -> Dict[str, Any]:
    """Reads and returns the contents of the config file."""
    # Get the directory of the current script
    script_dir = os.path.dirname(__file__)
    # Build the path to the config file
    config_file_path = os.path.join(script_dir, "config.json")
    # Load the config file
    with open(config_file_path, "r", encoding="utf-8") as file:
        config = json.load(file)
    return config



"""This module implements visualization for the Pose Estimation task."""

import os
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np

from senselab.video.data_structures.pose import ImagePose
from senselab.video.tasks.pose_estimation.utils import SENSELAB_KEYPOINT_MAPPING

try:
    import cv2

    CV2_AVAILABLE = True
except ModuleNotFoundError:
    CV2_AVAILABLE = False

try:
    from mediapipe import solutions
    from mediapipe.framework.formats import landmark_pb2

    MEDIAPIPE_AVAILABLE = True
except ModuleNotFoundError:
    MEDIAPIPE_AVAILABLE = False


def visualize(pose_image: ImagePose, output_path: Optional[str] = None, plot: bool = False) -> np.ndarray:
    """Visualize detected poses by drawing landmarks and connections on the image.

    Args:
        pose_image (ImagePose): The pose estimation result containing detected poses.
        output_path (str): Optional path to save the visualized image. If provided, saves the
            annotated image to this location.
        plot (bool): Whether to display the annotated image using matplotlib.

    Returns:
        np.ndarray: The input image with pose landmarks and connections drawn on it.
    """
    if not CV2_AVAILABLE:
        raise ModuleNotFoundError(
            "`opencv-python` is not installed. "
            "Please install senselab video dependencies using `pip install senselab['video']`."
        )
    if not MEDIAPIPE_AVAILABLE:
        raise ModuleNotFoundError(
            "`mediapipe` is not installed. "
            "Please install senselab video dependencies using `pip install senselab['video']`."
        )

    annotated_image = pose_image.image.copy()

    for individual in pose_image.individuals:
        pose_landmarks_proto = landmark_pb2.NormalizedLandmarkList()
        landmarks = []
        pose_lm = individual.normalized_landmarks
        # Filter out landmarks with low confidence
        landmarks = [
            landmark_pb2.NormalizedLandmark(
                x=getattr(pose_lm[lm], "x", 0), y=getattr(pose_lm[lm], "y", 0), z=getattr(pose_lm[lm], "z", 0)
            )  # type: ignore[attr-defined]
            if (lm in pose_lm and getattr(pose_lm[lm], "confidence", 1) > 0.5)
            else landmark_pb2.NormalizedLandmark(x=0, y=0, z=0, visibility=0)
            for lm in SENSELAB_KEYPOINT_MAPPING.values()
        ]
        pose_landmarks_proto.landmark.extend(landmarks)
        solutions.drawing_utils.draw_landmarks(
            annotated_image,
            pose_landmarks_proto,
            solutions.pose.POSE_CONNECTIONS,
            solutions.drawing_styles.get_default_pose_landmarks_style(),
        )

    if output_path:
        print(f"Saving visualization to {output_path}")
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        cv2.imwrite(output_path, cv2.cvtColor(annotated_image, cv2.COLOR_RGB2BGR))

    if plot:
        plt.imshow(annotated_image)
        plt.axis("off")
        plt.show()
        plt.close()

    return annotated_image



"""This module provides the implementation of torchaudio squim utilities for audio features extraction."""

from typing import Any, Dict, List

import numpy as np

from senselab.audio.data_structures import Audio

try:
    from torchaudio.pipelines import SQUIM_OBJECTIVE, SQUIM_SUBJECTIVE

    objective_model = SQUIM_OBJECTIVE.get_model()
    subjective_model = SQUIM_SUBJECTIVE.get_model()
    TORCHAUDIO_AVAILABLE = True
except ModuleNotFoundError:
    TORCHAUDIO_AVAILABLE = False


def extract_objective_quality_features_from_audios(audios: List["Audio"]) -> List[Dict[str, Any]]:
    """Extracts objective audio features from a list of Audio objects.

    Features include:
    - Wideband Perceptual Estimation of Speech Quality (PESQ)
    - Short-Time Objective Intelligibility (STOI)
    - Scale-Invariant Signal-to-Distortion Ratio (SI-SDR)

    Currently, Torchaudio-Squim model only supports mono audio at 16000 Hz sampling rate.

    Args:
        audios (List[Audio]): List of Audio objects.

    Returns:
        List[Dict[str, Any]]: List of dictionaries, each containing extracted features for an audio input.
    """
    if not TORCHAUDIO_AVAILABLE:
        raise ModuleNotFoundError(
            "`torchaudio` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )

    if any(audio.waveform.shape[0] != 1 for audio in audios):
        raise ValueError("Only mono audio is supported by Torchaudio-Squim model.")

    if any(audio.sampling_rate != 16000 for audio in audios):
        raise ValueError("Only 16000 Hz sampling rate is supported by Torchaudio-Squim model.")

    features: List[Dict[str, Any]] = []

    for audio in audios:
        audio_features = {}
        try:
            stoi, pesq, si_sdr = objective_model(audio.waveform)
            audio_features["stoi"] = stoi.item()
            audio_features["pesq"] = pesq.item()
            audio_features["si_sdr"] = si_sdr.item()
        except RuntimeError:
            audio_features["stoi"] = np.nan
            audio_features["pesq"] = np.nan
            audio_features["si_sdr"] = np.nan

        features.append(audio_features)

    return features


def extract_subjective_quality_features_from_audios(
    audios: List["Audio"], non_matching_references: List["Audio"]
) -> List[Dict[str, Any]]:
    """Extracts subjective audio features from a list of Audio objects.

    Currently, Torchaudio-Squim model only supports mono audio at 16000 Hz sampling rate.

    Args:
        audios (List[Audio]): List of Audio objects.
        non_matching_references (List[Audio]): Reference Audio objects for the subjective model.

    Returns:
        List[Dict[str, Any]]: List of dictionaries, each containing extracted features for an audio input.
    """
    if not TORCHAUDIO_AVAILABLE:
        raise ModuleNotFoundError(
            "`torchaudio` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )

    # Check if any audio is not mono
    if any(audio.waveform.shape[0] != 1 for audio in audios) or any(
        ref.waveform.shape[0] != 1 for ref in non_matching_references
    ):
        raise ValueError("Only mono audio is supported by Torchaudio-Squim model.")

    # Check if any audio has a sampling rate other than 16000 Hz
    if any(audio.sampling_rate != 16000 for audio in audios) or any(
        ref.sampling_rate != 16000 for ref in non_matching_references
    ):
        raise ValueError("Only 16000 Hz sampling rate is supported by Torchaudio-Squim model.")

    features: List[Dict[str, Any]] = []

    for i, audio in enumerate(audios):
        audio_features = {}
        try:
            mos = subjective_model(audio.waveform, non_matching_references[i].waveform)
            audio_features["mos"] = mos.item()
        except RuntimeError:
            audio_features["mos"] = np.nan

        features.append(audio_features)

    return features



"""This module provides utilities for pose estimation."""

import os

import requests

# Path where models are stored
MODEL_PATH = "src/senselab/video/tasks/pose_estimation/models"

# Senselab Pose Estimation keypoint mapping
SENSELAB_KEYPOINT_MAPPING = {
    0: "Nose",
    1: "Left Eye Inner",
    2: "Left Eye",
    3: "Left Eye Outer",
    4: "Right Eye Inner",
    5: "Right Eye",
    6: "Right Eye Outer",
    7: "Left Ear",
    8: "Right Ear",
    9: "Mouth Left",
    10: "Mouth Right",
    11: "Left Shoulder",
    12: "Right Shoulder",
    13: "Left Elbow",
    14: "Right Elbow",
    15: "Left Wrist",
    16: "Right Wrist",
    17: "Left Pinky",
    18: "Right Pinky",
    19: "Left Index",
    20: "Right Index",
    21: "Left Thumb",
    22: "Right Thumb",
    23: "Left Hip",
    24: "Right Hip",
    25: "Left Knee",
    26: "Right Knee",
    27: "Left Ankle",
    28: "Right Ankle",
    29: "Left Heel",
    30: "Right Heel",
    31: "Left Foot Index",
    32: "Right Foot Index",
}

YOLO_KEYPOINT_MAPPING = {
    0: SENSELAB_KEYPOINT_MAPPING[0],  # Nose
    1: SENSELAB_KEYPOINT_MAPPING[2],  # Left Eye
    2: SENSELAB_KEYPOINT_MAPPING[5],  # Right Eye
    3: SENSELAB_KEYPOINT_MAPPING[7],  # Left Ear
    4: SENSELAB_KEYPOINT_MAPPING[8],  # Right Ear
    5: SENSELAB_KEYPOINT_MAPPING[11],  # Left Shoulder
    6: SENSELAB_KEYPOINT_MAPPING[12],  # Right Shoulder
    7: SENSELAB_KEYPOINT_MAPPING[13],  # Left Elbow
    8: SENSELAB_KEYPOINT_MAPPING[14],  # Right Elbow
    9: SENSELAB_KEYPOINT_MAPPING[15],  # Left Wrist
    10: SENSELAB_KEYPOINT_MAPPING[16],  # Right Wrist
    11: SENSELAB_KEYPOINT_MAPPING[23],  # Left Hip
    12: SENSELAB_KEYPOINT_MAPPING[24],  # Right Hip
    13: SENSELAB_KEYPOINT_MAPPING[25],  # Left Knee
    14: SENSELAB_KEYPOINT_MAPPING[26],  # Right Knee
    15: SENSELAB_KEYPOINT_MAPPING[27],  # Left Ankle
    16: SENSELAB_KEYPOINT_MAPPING[28],  # Right Ankle
}

MEDIAPIPE_KEYPOINT_MAPPING = {
    0: SENSELAB_KEYPOINT_MAPPING[0],  # Nose
    1: SENSELAB_KEYPOINT_MAPPING[1],  # Left Eye Inner
    2: SENSELAB_KEYPOINT_MAPPING[2],  # Left Eye
    3: SENSELAB_KEYPOINT_MAPPING[3],  # Left Eye Outer
    4: SENSELAB_KEYPOINT_MAPPING[4],  # Right Eye Inner
    5: SENSELAB_KEYPOINT_MAPPING[5],  # Right Eye
    6: SENSELAB_KEYPOINT_MAPPING[6],  # Right Eye Outer
    7: SENSELAB_KEYPOINT_MAPPING[7],  # Left Ear
    8: SENSELAB_KEYPOINT_MAPPING[8],  # Right Ear
    9: SENSELAB_KEYPOINT_MAPPING[9],  # Mouth Left
    10: SENSELAB_KEYPOINT_MAPPING[10],  # Mouth Right
    11: SENSELAB_KEYPOINT_MAPPING[11],  # Left Shoulder
    12: SENSELAB_KEYPOINT_MAPPING[12],  # Right Shoulder
    13: SENSELAB_KEYPOINT_MAPPING[13],  # Left Elbow
    14: SENSELAB_KEYPOINT_MAPPING[14],  # Right Elbow
    15: SENSELAB_KEYPOINT_MAPPING[15],  # Left Wrist
    16: SENSELAB_KEYPOINT_MAPPING[16],  # Right Wrist
    17: SENSELAB_KEYPOINT_MAPPING[17],  # Left Pinky
    18: SENSELAB_KEYPOINT_MAPPING[18],  # Right Pinky
    19: SENSELAB_KEYPOINT_MAPPING[19],  # Left Index
    20: SENSELAB_KEYPOINT_MAPPING[20],  # Right Index
    21: SENSELAB_KEYPOINT_MAPPING[21],  # Left Thumb
    22: SENSELAB_KEYPOINT_MAPPING[22],  # Right Thumb
    23: SENSELAB_KEYPOINT_MAPPING[23],  # Left Hip
    24: SENSELAB_KEYPOINT_MAPPING[24],  # Right Hip
    25: SENSELAB_KEYPOINT_MAPPING[25],  # Left Knee
    26: SENSELAB_KEYPOINT_MAPPING[26],  # Right Knee
    27: SENSELAB_KEYPOINT_MAPPING[27],  # Left Ankle
    28: SENSELAB_KEYPOINT_MAPPING[28],  # Right Ankle
    29: SENSELAB_KEYPOINT_MAPPING[29],  # Left Heel
    30: SENSELAB_KEYPOINT_MAPPING[30],  # Right Heel
    31: SENSELAB_KEYPOINT_MAPPING[31],  # Left Foot Index
    32: SENSELAB_KEYPOINT_MAPPING[32],  # Right Foot Index
}

# Available models
MODELS = {
    "mediapipe": {
        "lite": "https://storage.googleapis.com/mediapipe-models/pose_landmarker/pose_landmarker_lite/float16/1/pose_landmarker_lite.task",
        "full": "https://storage.googleapis.com/mediapipe-models/pose_landmarker/pose_landmarker_full/float16/1/pose_landmarker_full.task",
        "heavy": "https://storage.googleapis.com/mediapipe-models/pose_landmarker/pose_landmarker_heavy/float16/1/pose_landmarker_heavy.task",
    },
    "yolo": {
        "8n": "yolov8n-pose.pt",
        "8s": "yolov8s-pose.pt",
        "8m": "yolov8m-pose.pt",
        "8l": "yolov8l-pose.pt",
        "8x": "yolov8x-pose.pt",
        "11n": "yolo11n-pose.pt",
        "11s": "yolo11s-pose.pt",
        "11m": "yolo11m-pose.pt",
        "11l": "yolo11l-pose.pt",
        "11x": "yolo11x-pose.pt",
    },
}


def get_model(model: str, model_type: str) -> str:
    """Retrieve the model file or name, depending on the model type.

    Args:
        model (str): The model category ('mediapipe' or 'yolo').
        model_type (str): The specific model type (e.g., 'lite', 'full', '8n').

    Returns:
        str: The local path to the MediaPipe model file or the YOLO model name.

    Raises:
        ValueError: If an invalid model type is provided.
    """
    if model not in MODELS:
        raise ValueError(f"Invalid model category '{model}'. Choose from {list(MODELS.keys())}.")

    if model_type not in MODELS[model]:
        raise ValueError(f"Invalid {model} model type '{model_type}'. Choose from {list(MODELS[model].keys())}.")

    model_url = MODELS[model][model_type]
    cache_dir = os.path.join(MODEL_PATH, model)
    os.makedirs(cache_dir, exist_ok=True)

    if model == "mediapipe":
        # For MediaPipe, download the model file if necessary
        model_filename = os.path.basename(model_url)
        model_path = os.path.join(cache_dir, model_filename)
        download_model(model_url, model_path)
        return model_path

    elif model == "yolo":
        # For YOLO, just return the model path
        model_path = os.path.join(cache_dir, model_url)
        return model_path

    return ""


def download_model(url: str, save_path: str) -> None:
    """Download a model file from a URL if it doesn't already exist locally.

    Args:
        url (str): The URL of the model file.
        save_path (str): The local path to save the model file.
    """
    if not os.path.exists(save_path):
        print(f"Downloading model from {url}...")
        response = requests.get(url, stream=True)
        response.raise_for_status()
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        with open(save_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        print(f"Model saved to {save_path}")
    else:
        print(f"Model already exists at {save_path}")



"""This module implements some utilities for the text-to-speech task."""

from typing import Any, List, Optional, Tuple, TypeGuard

from senselab.audio.data_structures import Audio
from senselab.audio.tasks.text_to_speech.huggingface import HuggingFaceTTS
from senselab.audio.tasks.text_to_speech.marstts import Mars5TTS
from senselab.utils.data_structures import DeviceType, HFModel, Language, SenselabModel, TorchModel


def synthesize_texts(
    texts: List[str],
    model: Optional[SenselabModel] = None,
    language: Optional[Language] = None,
    device: Optional[DeviceType] = None,
    targets: Optional[List[Audio | Tuple[Audio, str]]] = None,
    **kwargs: Any,  # noqa: ANN401
) -> List[Audio]:
    """Synthesizes speech from all texts using the given model.

    This function synthesizes speech from a list of text strings using the specified text-to-speech (TTS) model.
    It supports models from HuggingFace and `Mars5TTS` for now.

    Args:
        texts (List[str]): The list of text strings to be synthesized.
        model (SenselabModel): The model used for synthesis.
                If None, the default model "suno/bark" is used.
        language (Optional[Language]): The language of the text
            (default is None).
        device (Optional[DeviceType]): The device to run the model on
            (default is None).
        targets (Optional[List[Audio | Tuple[Audio, str]]]):
            A list where each element is a target audio or a tuple of target audio and transcript.
            Depending on the model being used, the `target` input may need to be provided in a specific format:
            - Hugging Face models do not require a `target` input at all.
            - `Mars5TTS` requires both `Audio` and a transcript for all inputs.
        **kwargs: Additional keyword arguments to pass to the synthesis function.
            Depending on the model used (e.g., HFModel), additional arguments
            may be required. Refer to the model-specific documentation for details.

    Returns:
        List[Audio]: The list of synthesized audio objects.
    """
    if model is None:
        model = HFModel(path_or_uri="suno/bark", revision="main")

    if targets is not None:
        assert len(targets) == len(texts), ValueError("Provided targets should be same length as texts")

        for i, target in targets:
            if isinstance(target, tuple):
                assert len(target[1]) > 0, ValueError(f"{i}th target was expected to have a transcript, but was empty.")

    if isinstance(model, HFModel):
        return HuggingFaceTTS.synthesize_texts_with_transformers(texts=texts, model=model, device=device, **kwargs)
    elif isinstance(model, TorchModel):
        if model.path_or_uri == "Camb-ai/mars5-tts":
            # Converting target to the required format for Mars5TTS
            assert targets is not None, ValueError(
                "Mars5-TTS requires target audios and their corresponding transcripts."
            )

            if _check_all_have_transcripts(targets):
                return Mars5TTS.synthesize_texts_with_mars5tts(
                    texts=texts, targets=targets, language=language, device=device, **kwargs
                )
            else:
                raise ValueError("Mars5-TTS requires target audios and their corresponding transcripts.")
        else:
            raise NotImplementedError(f"{model.path_or_uri} is currently not a supported Torch model. \
                                      Feel free to reach out to us about integrating this model into senselab.")
    else:
        raise NotImplementedError("Only Hugging Face models and select Torch models are supported for now.")


def _check_all_have_transcripts(targets: List[Audio | Tuple[Audio, str]]) -> TypeGuard[List[Tuple[Audio, str]]]:
    for target in targets:
        if isinstance(target, Audio):
            return False
        elif isinstance(target, tuple):
            if len(target) != 2 or not isinstance(target[0], Audio) or not isinstance(target[1], str):
                return False
            elif len(target[1]) == 0:
                return False

    return True



"""This module implements some utilities for audio data augmentation with audiomentations."""

from typing import List

from senselab.audio.data_structures import Audio

try:
    from audiomentations import Compose

    AUDIOMENTATIONS_AVAILABLE = True
except ModuleNotFoundError:
    AUDIOMENTATIONS_AVAILABLE = False


def augment_audios_with_audiomentations(audios: List[Audio], augmentation: "Compose") -> List[Audio]:
    """Augments all provided audios with audiomentations library.

    Args:
        audios: List of Audios whose data will be augmented with the given augmentations.
        augmentation: A Composition of augmentations to run on each audio (uses audiomentations).

    Returns:
        List of audios that have passed through the provided augmentation.
    """

    def _augment_single_audio(audio: Audio, augmentation: "Compose"):  # noqa: ANN202
        """Augments a single audio with audiomentations.

        Args:
            audio: The audio to be augmented.
            augmentation: The audiomentations augmentation to be applied.

        Returns:
            The augmented audio. The returned data type is not explicitly specified
                in the function signature because it would brake pydra.
        """
        augmented_waveform = augmentation(samples=audio.waveform.numpy(), sample_rate=audio.sampling_rate)
        return Audio(
            waveform=augmented_waveform,
            sampling_rate=audio.sampling_rate,
            metadata=audio.metadata.copy(),
        )

    if not AUDIOMENTATIONS_AVAILABLE:
        raise ModuleNotFoundError(
            "`audiomentations` is not installed. "
            "Please install senselab audio dependencies using `pip install senselab['audio']`."
        )

    """
    # The commented code is for parallelizing the augmentation using pydra
    # Due to some issues with pydra, this is disabled for now

    import pydra

    _augment_single_audio_pt = pydra.mark.task(_augment_single_audio)

    # Create the workflow
    wf = pydra.Workflow(name="audio_augmentation_wf", input_spec=["y"])
    wf.split("y", y=audios)
    wf.add(_augment_single_audio_pt(name="augment_audio", audio=wf.lzin.y, augmentation=augmentation))
    wf.set_output([("augmented_audio", wf.augment_audio.lzout.out)])

    # Execute the workflow
    with pydra.Submitter(plugin="cf") as submitter:
        submitter(wf)

    outputs = wf.result()
    return [out.output.augmented_audio for out in outputs]
    """

    augmented_audios = []
    for audio in audios:
        augmented_audios.append(_augment_single_audio(audio, augmentation))

    return augmented_audios



"""This module provides the implementation of cosine similarity."""

import torch


def compute_cosine_similarity(tensor1: torch.Tensor, tensor2: torch.Tensor) -> float:
    """Compute the cosine similarity between two torch tensors.

    Args:
        tensor1 (Tensor): The first input tensor.
        tensor2 (Tensor): The second input tensor.

    Returns:
        float: The cosine similarity between the two input tensors.

    Raises:
        ValueError: If the input tensors are not of the same shape.

    Examples:
        >>> tensor1 = torch.tensor([1.0, 2.0, 3.0])
        >>> tensor2 = torch.tensor([4.0, 5.0, 6.0])
        >>> cosine_similarity(tensor1, tensor2)
        0.9746318461970762

        >>> tensor1 = torch.tensor([1.0, 0.0, -1.0])
        >>> tensor2 = torch.tensor([-1.0, 0.0, 1.0])
        >>> cosine_similarity(tensor1, tensor2)
        -1.0

    Note:
        This function assumes the input tensors are 1-dimensional and have the same shape.
    """
    if tensor1.dim() != 1 or tensor2.dim() != 1:
        raise ValueError("Input tensors must be 1-dimensional")
    if tensor1.shape != tensor2.shape:
        raise ValueError("Input tensors must have the same shape")

    dot_product = torch.dot(tensor1, tensor2)
    norm_tensor1 = torch.norm(tensor1)
    norm_tensor2 = torch.norm(tensor2)

    cosine_sim = dot_product / (norm_tensor1 * norm_tensor2)
    return cosine_sim.item()
